import logging
import re

from slither.solc_parsing.declarations.caller_context import CallerContextExpression
from slither.solc_parsing.expressions.expression_parsing import parse_expression

from slither.core.variables.variable import Variable

from slither.solc_parsing.solidity_types.type_parsing import parse_type, UnknownType

from slither.core.solidity_types.elementary_type import (
    ElementaryType,
    NonElementaryType,
)
from slither.solc_parsing.exceptions import ParsingError

logger = logging.getLogger("VariableDeclarationSolcParsing")


class MultipleVariablesDeclaration(Exception):
    """
    This is raised on
    var (a,b) = ...
    It should occur only on local variable definition
    """

    pass


class VariableDeclarationSolc:
    def __init__(self, variable: Variable, variable_data: dict) -> None:
        """
        A variable can be declared through a statement, or directly.
        If it is through a statement, the following children may contain
        the init value.
        It may be possible that the variable is declared through a statement,
        but the init value is declared at the VariableDeclaration children level
        """

        self._variable = variable
        self._was_analyzed = False
        self._elem_to_parse: dict | UnknownType | None = None
        self._initializedNotParsed: dict | None = None

        self._is_compact_ast = False

        self._reference_id: int | None = None

        if "nodeType" in variable_data:
            self._is_compact_ast = True
            nodeType = variable_data["nodeType"]
            if nodeType in [
                "VariableDeclarationStatement",
                "VariableDefinitionStatement",
            ]:
                if len(variable_data["declarations"]) > 1:
                    raise MultipleVariablesDeclaration
                init = None
                if "initialValue" in variable_data:
                    init = variable_data["initialValue"]
                self._init_from_declaration(variable_data["declarations"][0], init)
            elif nodeType == "VariableDeclaration":
                self._init_from_declaration(variable_data, variable_data.get("value"))
            else:
                raise ParsingError(f"Incorrect variable declaration type {nodeType}")

        else:
            nodeType = variable_data["name"]

            if nodeType in [
                "VariableDeclarationStatement",
                "VariableDefinitionStatement",
            ]:
                if len(variable_data["children"]) == 2:
                    init = variable_data["children"][1]
                elif len(variable_data["children"]) == 1:
                    init = None
                elif len(variable_data["children"]) > 2:
                    raise MultipleVariablesDeclaration
                else:
                    raise ParsingError(
                        "Variable declaration without children?" + str(variable_data)
                    )
                declaration = variable_data["children"][0]
                self._init_from_declaration(declaration, init)
            elif nodeType == "VariableDeclaration":
                self._init_from_declaration(variable_data, None)
            else:
                raise ParsingError(f"Incorrect variable declaration type {nodeType}")

    @property
    def underlying_variable(self) -> Variable:
        return self._variable

    @property
    def reference_id(self) -> int:
        """
        Return the solc id. It can be compared with the referencedDeclaration attr
        Returns None if it was not parsed (legacy AST)
        """
        assert self._reference_id
        return self._reference_id

    def _handle_comment(self, attributes: dict) -> None:
        if "documentation" in attributes and "text" in attributes["documentation"]:
            candidates = attributes["documentation"]["text"].split(",")

            for candidate in candidates:
                if "@custom:security non-reentrant" in candidate:
                    self._variable.is_reentrant = False

                write_protection = re.search(
                    r'@custom:security write-protection="([\w, ()]*)"', candidate
                )
                if write_protection:
                    if self._variable.write_protection is None:
                        self._variable.write_protection = []
                    self._variable.write_protection.append(write_protection.group(1))

    def _analyze_variable_attributes(self, attributes: dict) -> None:
        self._variable.visibility = attributes.get("visibility", "internal")

    def _init_from_declaration(self, var: dict, init: dict | None) -> None:
        if self._is_compact_ast:
            attributes = var
            self._typeName = attributes["typeDescriptions"]["typeString"]
        else:
            assert len(var["children"]) <= 2
            assert var["name"] == "VariableDeclaration"

            attributes = var["attributes"]
            self._typeName = attributes["type"]

        self._variable.name = attributes["name"]
        # self._arrayDepth = 0
        # self._isMapping = False
        # self._mappingFrom = None
        # self._mappingTo = False
        # self._initial_expression = None
        # self._type = None

        # Only for comapct ast format
        # the id can be used later if referencedDeclaration
        # is provided
        if "id" in var:
            self._reference_id = var["id"]

        if "constant" in attributes:
            self._variable.is_constant = attributes["constant"]

        if "mutability" in attributes:
            # Note: this checked is not needed if "constant" was already in attribute, but we keep it
            # for completion
            if attributes["mutability"] == "constant":
                self._variable.is_constant = True
            if attributes["mutability"] == "immutable":
                self._variable.is_immutable = True

        self._handle_comment(attributes)

        self._analyze_variable_attributes(attributes)

        if self._is_compact_ast:
            if var["typeName"]:
                self._elem_to_parse = var["typeName"]
            else:
                self._elem_to_parse = UnknownType(var["typeDescriptions"]["typeString"])
        else:
            if not var["children"]:
                # It happens on variable declared inside loop declaration
                try:
                    self._variable.type = ElementaryType(self._typeName)
                    self._elem_to_parse = None
                except NonElementaryType:
                    self._elem_to_parse = UnknownType(self._typeName)
            else:
                self._elem_to_parse = var["children"][0]

        if self._is_compact_ast:
            self._initializedNotParsed = init
            if init:
                self._variable.initialized = True
        else:
            if init:  # there are two way to init a var local in the AST
                assert len(var["children"]) <= 1
                self._variable.initialized = True
                self._initializedNotParsed = init
            elif len(var["children"]) in [0, 1]:
                self._variable.initialized = False
                self._initializedNotParsed = None
            else:
                assert len(var["children"]) == 2
                self._variable.initialized = True
                self._initializedNotParsed = var["children"][1]

    def analyze(self, caller_context: CallerContextExpression) -> None:
        # Can be re-analyzed due to inheritance
        if self._was_analyzed:
            return
        self._was_analyzed = True

        if self._elem_to_parse:
            self._variable.type = parse_type(self._elem_to_parse, caller_context)
            self._elem_to_parse = None

        if self._variable.initialized:
            assert self._initializedNotParsed
            self._variable.expression = parse_expression(self._initializedNotParsed, caller_context)
            self._initializedNotParsed = None



"""
Various utils for sarif/vscode
"""

import json
from pathlib import Path
from typing import Any


def _parse_index(key: str) -> tuple[int, int] | None:
    if key.count(":") != 2:
        return None

    try:
        run = int(key[key.find(":") + 1 : key.rfind(":")])
        index = int(key[key.rfind(":") + 1 :])
        return run, index
    except ValueError:
        return None


def _get_indexes(path_to_triage: Path) -> list[tuple[int, int]]:
    try:
        with open(path_to_triage, encoding="utf8") as file_desc:
            triage = json.load(file_desc)
    except json.decoder.JSONDecodeError:
        return []

    resultIdToNotes: dict[str, dict] = triage.get("resultIdToNotes", {})

    indexes: list[tuple[int, int]] = []
    for key, data in resultIdToNotes.items():
        if "status" in data and data["status"] == 1:
            parsed = _parse_index(key)
            if parsed:
                indexes.append(parsed)

    return indexes


def read_triage_info(path_to_sarif: Path, path_to_triage: Path) -> list[str]:
    try:
        with open(path_to_sarif, encoding="utf8") as file_desc:
            sarif = json.load(file_desc)
    except json.decoder.JSONDecodeError:
        return []

    runs: list[dict[str, Any]] = sarif.get("runs", [])

    # Don't support multiple runs for now
    if len(runs) != 1:
        return []

    run_results: list[dict] = runs[0].get("results", [])

    indexes = _get_indexes(path_to_triage)

    ids: list[str] = []
    for run, index in indexes:
        # We dont support multiple runs for now
        if run != 0:
            continue
        try:
            elem = run_results[index]
        except KeyError:
            continue
        if "partialFingerprints" in elem:
            if "id" in elem["partialFingerprints"]:
                ids.append(elem["partialFingerprints"]["id"])

    return ids



"""
CK Metrics are a suite of six software metrics proposed by Chidamber and Kemerer in 1994.
These metrics are used to measure the complexity of a class.
https://en.wikipedia.org/wiki/Programming_complexity

- Response For a Class (RFC) is a metric that measures the number of unique method calls within a class.
- Number of Children (NOC) is a metric that measures the number of children a class has.
- Depth of Inheritance Tree (DIT) is a metric that measures the number of parent classes a class has.
- Coupling Between Object Classes (CBO) is a metric that measures the number of classes a class is coupled to.

Not implemented:
- Lack of Cohesion of Methods (LCOM) is a metric that measures the lack of cohesion in methods.
- Weighted Methods per Class (WMC) is a metric that measures the complexity of a class.

During the calculation of the metrics above, there are a number of other intermediate metrics that are calculated.
These are also included in the output:
 - State variables: total number of state variables
 - Constants: total number of constants
 - Immutables: total number of immutables
 - Public: total number of public functions
 - External: total number of external functions
 - Internal: total number of internal functions
 - Private: total number of private functions
 - Mutating: total number of state mutating functions
 - View: total number of view functions
 - Pure: total number of pure functions
 - External mutating: total number of external mutating functions
 - No auth or onlyOwner: total number of functions without auth or onlyOwner modifiers
 - No modifiers: total number of functions without modifiers
 - Ext calls: total number of external calls

"""

from slither.printers.abstract_printer import AbstractPrinter
from slither.utils.ck import CKMetrics
from slither.utils.output import Output


class CK(AbstractPrinter):
    ARGUMENT = "ck"
    HELP = "Chidamber and Kemerer (CK) complexity metrics and related function attributes"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#ck"

    def output(self, _filename: str) -> Output:
        if len(self.contracts) == 0:
            return self.generate_output("No contract found")

        ck = CKMetrics(self.contracts)

        res = self.generate_output(ck.full_text)
        res.add_pretty_table(ck.auxiliary1.pretty_table, ck.auxiliary1.title)
        res.add_pretty_table(ck.auxiliary2.pretty_table, ck.auxiliary2.title)
        res.add_pretty_table(ck.auxiliary3.pretty_table, ck.auxiliary3.title)
        res.add_pretty_table(ck.auxiliary4.pretty_table, ck.auxiliary4.title)
        res.add_pretty_table(ck.core.pretty_table, ck.core.title)
        self.info(ck.full_text)

        return res



from slither.core.source_mapping.source_mapping import SourceMapping


class Enum(SourceMapping):
    def __init__(self, name: str, canonical_name: str, values: list[str]) -> None:
        super().__init__()
        self._name = name
        self._canonical_name = canonical_name
        self._values = values
        self._min = 0
        # The max value of an Enum is the index of the last element
        self._max = len(values) - 1

    @property
    def canonical_name(self) -> str:
        return self._canonical_name

    @property
    def name(self) -> str:
        return self._name

    @property
    def values(self) -> list[str]:
        return self._values

    @property
    def min(self) -> int:
        return self._min

    @property
    def max(self) -> int:
        return self._max

    def __str__(self) -> str:
        return self.name



from slither.core.expressions.expression import Expression


class NewContract(Expression):
    def __init__(self, contract_name: str) -> None:
        super().__init__()
        self._contract_name: str = contract_name
        self._gas = None
        self._value = None
        self._salt = None

    @property
    def contract_name(self) -> str:
        return self._contract_name

    @property
    def call_value(self):
        return self._value

    @call_value.setter
    def call_value(self, v):
        self._value = v

    @property
    def call_salt(self):
        return self._salt

    @call_salt.setter
    def call_salt(self, salt):
        self._salt = salt

    def __str__(self) -> str:
        return "new " + str(self._contract_name)



from slither.core.variables.local_variable import LocalVariable


class LocalVariableInitFromTuple(LocalVariable):
    """
    Use on this pattern:
    var(a,b) = f()

    It is not possible to split the variable declaration in sigleton and keep the init value
    We init a and b with f(). get_tuple_index ret() returns which returns values of f is to be used

    """

    def __init__(self) -> None:
        super().__init__()
        self._tuple_index: int | None = None

    @property
    def tuple_index(self) -> int | None:
        return self._tuple_index

    @tuple_index.setter
    def tuple_index(self, idx: int):
        self._tuple_index = idx



from typing import TYPE_CHECKING, TypeAlias
from slither.core.solidity_types import Type, UserDefinedType

if TYPE_CHECKING:
    from slither.core.declarations import Function

USING_FOR_KEY: TypeAlias = "str | Type"  # "*" is wildcard
USING_FOR_ITEM: TypeAlias = "list[UserDefinedType | Function]"  # UserDefinedType.type is a library
USING_FOR: TypeAlias = "dict[USING_FOR_KEY, USING_FOR_ITEM]"


def merge_using_for(uf1: USING_FOR, uf2: USING_FOR) -> USING_FOR:
    result = {**uf1, **uf2}
    for key, value in result.items():
        if key in uf1 and key in uf2:
            result[key] = value + uf1[key]
    return result



from slither.slithir.operations.call import Call
from slither.core.variables.variable import Variable
from slither.core.declarations.solidity_variables import SolidityVariable
from slither.core.variables.local_variable import LocalVariable
from slither.slithir.variables.constant import Constant
from slither.slithir.variables.local_variable import LocalIRVariable


class Transfer(Call):
    def __init__(self, destination: LocalVariable | LocalIRVariable, value: Constant) -> None:
        assert isinstance(destination, (Variable, SolidityVariable))
        self._destination = destination
        super().__init__()

        self._call_value = value

    def can_send_eth(self) -> bool:
        return True

    @property
    def call_value(self) -> Constant:
        return self._call_value

    @property
    def read(self) -> list[Constant | LocalIRVariable | LocalVariable]:
        return [self.destination, self.call_value]

    @property
    def destination(self) -> LocalVariable | LocalIRVariable:
        return self._destination

    def __str__(self):
        value = f"value:{self.call_value}"
        return f"Transfer dest:{self.destination} {value}"



"""
Module detecting state changes in assert calls
"""

from slither.core.cfg.node import Node
from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations.internal_call import InternalCall
from slither.utils.output import Output


def detect_assert_state_change(
    contract: Contract,
) -> list[tuple[FunctionContract, Node]]:
    """
    Detects and returns all nodes with assert calls that change contract state from within the invariant
    :param contract: Contract to detect
    :return: A list of nodes with assert calls that change contract state from within the invariant
    """

    # Create our result set.
    # List of tuples (function, node)
    results = []

    # Loop for each function and modifier.
    for function in contract.functions_declared + list(contract.modifiers_declared):
        for ir_call in function.internal_calls:
            # Detect assert() calls
            if ir_call.function.name == "assert(bool)" and (
                # Detect direct changes to state
                ir_call.node.state_variables_written
                or
                # Detect changes to state via function calls
                any(
                    ir
                    for ir in ir_call.node.irs
                    if isinstance(ir, InternalCall)
                    and ir.function
                    and ir.function.state_variables_written
                )
            ):
                results.append((function, ir_call.node))

    # Return the resulting set of nodes
    return results


class AssertStateChange(AbstractDetector):
    """
    Assert state change
    """

    ARGUMENT = "assert-state-change"
    HELP = "Assert state change"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#assert-state-change"
    WIKI_TITLE = "Assert state change"
    WIKI_DESCRIPTION = """Incorrect use of `assert()`. See Solidity best [practices](https://solidity.readthedocs.io/en/latest/control-structures.html#id4)."""

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract A {

  uint s_a;

  function bad() public {
    assert((s_a += 1) > 10);
  }
}
```
The assert in `bad()` increments the state variable `s_a` while checking for the condition.
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = """Use `require` for invariants modifying the state."""

    def _detect(self) -> list[Output]:
        """
        Detect assert calls that change state from within the invariant
        """
        results = []
        for contract in self.contracts:
            assert_state_change = detect_assert_state_change(contract)
            for func, node in assert_state_change:
                info: DETECTOR_INFO = [
                    func,
                    " has an assert() call which possibly changes state.\n",
                ]
                info += ["\t-", node, "\n"]
                info += [
                    "Consider using require() or change the invariant to not modify the state.\n"
                ]
                res = self.generate_result(info)
                results.append(res)
        return results



"""
Module detecting unused return values from low level
"""

from slither.core.cfg.node import Node
from slither.slithir.operations import LowLevelCall

from slither.core.declarations.function_contract import FunctionContract
from slither.core.variables.state_variable import StateVariable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class UncheckedLowLevel(AbstractDetector):
    """
    If the return value of a low-level call is not checked, it might lead to losing ether
    """

    ARGUMENT = "unchecked-lowlevel"
    HELP = "Unchecked low-level calls"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#unchecked-low-level-calls"

    WIKI_TITLE = "Unchecked low-level calls"
    WIKI_DESCRIPTION = "The return value of a low-level call is not checked."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract MyConc{
    function my_func(address payable dst) public payable{
        dst.call.value(msg.value)("");
    }
}
```
The return value of the low-level call is not checked, so if the call fails, the Ether will be locked in the contract.
If the low level is used to prevent blocking operations, consider logging failed calls.
    """
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Ensure that the return value of a low-level call is checked or logged."

    @staticmethod
    def detect_unused_return_values(f: FunctionContract) -> list[Node]:
        """
        Return the nodes where the return value of a call is unused
        Args:
            f (Function)
        Returns:
            list(Node)
        """
        values_returned = []
        nodes_origin = {}
        for n in f.nodes:
            for ir in n.irs:
                if isinstance(ir, LowLevelCall):
                    # if a return value is stored in a state variable, it's ok
                    if ir.lvalue and not isinstance(ir.lvalue, StateVariable):
                        values_returned.append(ir.lvalue)
                        nodes_origin[ir.lvalue] = ir

                for read in ir.read:
                    if read in values_returned:
                        values_returned.remove(read)

        return [nodes_origin[value].node for value in values_returned]

    def _detect(self) -> list[Output]:
        """Detect low level calls where the success value is not checked"""
        results = []
        for c in self.compilation_unit.contracts_derived:
            for f in c.functions_and_modifiers:
                unused_return = UncheckedLowLevel.detect_unused_return_values(f)
                if unused_return:
                    for node in unused_return:
                        info: DETECTOR_INFO = [f, " ignores return value by ", node, "\n"]

                        res = self.generate_result(info)

                        results.append(res)

        return results



import inspect
from slither import Slither
from slither.detectors import all_detectors
from slither.detectors.abstract_detector import AbstractDetector


def _run_all_detectors(slither: Slither) -> None:
    detectors = [getattr(all_detectors, name) for name in dir(all_detectors)]
    detectors = [d for d in detectors if inspect.isclass(d) and issubclass(d, AbstractDetector)]

    for detector in detectors:
        slither.register_detector(detector)

    slither.run_detectors()



from pathlib import Path
import pytest
from slither import Slither
from slither.core.declarations import Function, CustomErrorTopLevel, EventTopLevel
from slither.core.solidity_types.type_alias import TypeAliasTopLevel, TypeAliasContract
from slither.core.variables.top_level_variable import TopLevelVariable

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"
SRC_MAPPING_TEST_ROOT = Path(TEST_DATA_DIR, "src_mapping")


# Ensure issue fixed in https://github.com/crytic/crytic-compile/pull/554 does not regress in Slither's reference lookup.
@pytest.mark.parametrize("solc_version", ["0.6.12", "0.8.7", "0.8.8"])
def test_source_mapping_inheritance(solc_binary_path, solc_version):
    solc_path = solc_binary_path(solc_version)
    file = Path(SRC_MAPPING_TEST_ROOT, "inheritance.sol").as_posix()
    slither = Slither(file, solc=solc_path)

    # 3 reference to A in inheritance `contract $ is A`
    assert {(x.start, x.end) for x in slither.offset_to_references(file, 9)} == {
        (121, 122),
        (185, 186),
        (299, 300),
    }

    # Check if A.f() is at the offset 27
    functions = slither.offset_to_objects(file, 27)
    assert len(functions) == 1
    function = functions.pop()
    assert isinstance(function, Function)
    assert function.canonical_name == "A.f()"

    # Only one definition for A.f()
    assert {(x.start, x.end) for x in slither.offset_to_definitions(file, 27)} == {(26, 28)}
    # Only one reference for A.f(), in A.test()
    assert {(x.start, x.end) for x in slither.offset_to_references(file, 27)} == {(92, 93)}
    # Three overridden implementation of A.f(), in A.test()
    assert {(x.start, x.end) for x in slither.offset_to_implementations(file, 27)} == {
        (17, 53),
        (129, 166),
        (193, 230),
    }

    # Check if C.f() is at the offset 203
    functions = slither.offset_to_objects(file, 203)
    assert len(functions) == 1
    function = functions.pop()
    assert isinstance(function, Function)
    assert function.canonical_name == "C.f()"

    # Only one definition for C.f()
    assert {(x.start, x.end) for x in slither.offset_to_definitions(file, 203)} == {(202, 204)}
    # Two references for C.f(), in A.test() and C.test2()
    assert {(x.start, x.end) for x in slither.offset_to_references(file, 203)} == {
        (270, 271),
        (92, 93),
    }
    # Only one implementation for A.f(), in A.test()
    assert {(x.start, x.end) for x in slither.offset_to_implementations(file, 203)} == {(193, 230)}

    # Offset 93 is the call to f() in A.test()
    # This can lead to three different functions, depending on the current contract's context
    functions = slither.offset_to_objects(file, 93)
    print(functions)
    assert len(functions) == 3
    for function in functions:
        assert isinstance(function, Function)
        assert function.canonical_name in ["A.f()", "B.f()", "C.f()"]

    # There is one definition in the lexical scope of A
    assert {(x.start, x.end) for x in slither.offset_to_definitions(file, 93)} == {
        (26, 28),
    }

    # There are two references possible (in A.test() or C.test2() )
    assert {(x.start, x.end) for x in slither.offset_to_references(file, 93)} == {
        (92, 93),
        (270, 271),
    }

    # There are three implementations possible (in A, B or C)
    assert {(x.start, x.end) for x in slither.offset_to_implementations(file, 93)} == {
        (17, 53),
        (193, 230),
        (129, 166),
    }


def _sort_references_lines(refs: list) -> list:
    return sorted([ref.lines[0] for ref in refs])


def test_references_user_defined_aliases(solc_binary_path):
    """
    Tests if references are filled correctly for user defined aliases (declared using "type [...] is [...]" statement).
    """
    solc_path = solc_binary_path("0.8.16")
    file = Path(SRC_MAPPING_TEST_ROOT, "ReferencesUserDefinedAliases.sol").as_posix()
    slither = Slither(file, solc=solc_path)

    alias_top_level = slither.compilation_units[0].type_aliases["aliasTopLevel"]
    assert len(alias_top_level.references) == 2
    lines = _sort_references_lines(alias_top_level.references)
    assert lines == [12, 16]

    alias_contract_level = (
        slither.compilation_units[0].contracts[0].file_scope.type_aliases["C.aliasContractLevel"]
    )
    assert len(alias_contract_level.references) == 2
    lines = _sort_references_lines(alias_contract_level.references)
    assert lines == [13, 16]


def test_references_user_defined_types_when_casting(solc_binary_path):
    """
    Tests if references are filled correctly for user defined types in case of casting.
    """
    solc_path = solc_binary_path("0.8.16")
    file = Path(SRC_MAPPING_TEST_ROOT, "ReferencesUserDefinedTypesCasting.sol").as_posix()
    slither = Slither(file, solc=solc_path)

    contracts = slither.compilation_units[0].contracts
    a = contracts[0] if contracts[0].is_interface else contracts[1]
    assert len(a.references) == 2
    lines = _sort_references_lines(a.references)
    assert lines == [12, 18]


def test_source_mapping_top_level_defs(solc_binary_path):
    solc_path = solc_binary_path("0.8.24")
    file = Path(SRC_MAPPING_TEST_ROOT, "TopLevelReferences.sol").as_posix()
    slither = Slither(file, solc=solc_path)

    # Check if T is at the offset 5
    types = slither.offset_to_objects(file, 5)
    assert len(types) == 1
    type_ = types.pop()
    assert isinstance(type_, TypeAliasTopLevel)
    assert type_.name == "T"

    assert {(x.start, x.end) for x in slither.offset_to_references(file, 5)} == {
        (48, 49),
        (60, 61),
        (134, 135),
        (140, 141),
        (163, 164),
    }

    # Check if U is at the offset 33
    constants = slither.offset_to_objects(file, 33)
    assert len(constants) == 1
    constant = constants.pop()
    assert isinstance(constant, TopLevelVariable)
    assert constant.name == "U"
    assert {(x.start, x.end) for x in slither.offset_to_references(file, 33)} == {(147, 148)}

    # Check if V is at the offset 46
    errors = slither.offset_to_objects(file, 46)
    assert len(errors) == 1
    error = errors.pop()
    assert isinstance(error, CustomErrorTopLevel)
    assert error.name == "V"
    assert {(x.start, x.end) for x in slither.offset_to_references(file, 46)} == {(202, 203)}

    # Check if W is at the offset 58
    events = slither.offset_to_objects(file, 58)
    assert len(events) == 1
    event = events.pop()
    assert isinstance(event, EventTopLevel)
    assert event.name == "W"
    assert {(x.start, x.end) for x in slither.offset_to_references(file, 58)} == {(231, 232)}

    # Check if X is at the offset 87
    types = slither.offset_to_objects(file, 87)
    assert len(types) == 1
    type_ = types.pop()
    assert isinstance(type_, TypeAliasContract)
    assert type_.name == "X"
    assert {(x.start, x.end) for x in slither.offset_to_references(file, 87)} == {
        (245, 246),
        (251, 252),
    }


def test_references_self_identifier():
    """
    Tests that shadowing state variables with local variables does not affect references.
    """
    file = Path(SRC_MAPPING_TEST_ROOT, "SelfIdentifier.vy").as_posix()
    slither = Slither(file)

    contracts = slither.compilation_units[0].contracts
    a = contracts[0].state_variables[0]
    assert len(a.references) == 1
    lines = _sort_references_lines(a.references)
    assert lines == [4]



from slither.core.cfg.node import NodeType
from slither.tools.mutator.utils.patch import create_patch_with_line
from slither.tools.mutator.mutators.abstract_mutator import AbstractMutator
from slither.core.expressions.unary_operation import UnaryOperationType, UnaryOperation


class MIA(AbstractMutator):
    NAME = "MIA"
    HELP = '"if" construct around statement'

    def _mutate(self) -> dict:
        result: dict = {}
        for function in self.contract.functions_and_modifiers_declared:
            if not self.should_mutate_function(function):
                continue
            for node in function.nodes:
                if not self.should_mutate_node(node):
                    continue
                if node.type == NodeType.IF:
                    # Get the string
                    start = node.expression.source_mapping.start
                    stop = start + node.expression.source_mapping.length
                    old_str = node.source_mapping.content
                    line_no = node.source_mapping.lines
                    # Replace the expression with true and false
                    for value in ["true", "false"]:
                        new_str = value
                        create_patch_with_line(
                            result,
                            self.in_file,
                            start,
                            stop,
                            old_str,
                            new_str,
                            line_no[0],
                        )

                    if not isinstance(node.expression, UnaryOperation):
                        new_str = str(UnaryOperationType.BANG) + "(" + old_str + ")"
                        create_patch_with_line(
                            result,
                            self.in_file,
                            start,
                            stop,
                            old_str,
                            new_str,
                            line_no[0],
                        )
        return result



from pathlib import Path

from crytic_compile import CryticCompile
from crytic_compile.platform.solc_standard_json import SolcStandardJson
from slither import Slither

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data" / "inheritance_resolution"


# https://github.com/crytic/slither/issues/2304
def test_inheritance_with_renaming(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.0")
    standard_json = SolcStandardJson()
    for source_file in Path(TEST_DATA_DIR / "renaming").rglob("*.sol"):
        standard_json.add_source_file(Path(source_file).as_posix())
    compilation = CryticCompile(standard_json, solc=solc_path)
    slither = Slither(compilation)

    a = slither.get_contract_from_name("A")[0]
    b = slither.get_contract_from_name("B")[0]
    c = slither.get_contract_from_name("C")[0]

    assert len(a.immediate_inheritance) == 1
    assert a.immediate_inheritance[0] == b
    assert len(a.inheritance) == 2
    assert a.inheritance[0] == b
    assert a.inheritance[1] == c
    assert len(a.explicit_base_constructor_calls) == 1
    a_base_constructor_call = a.explicit_base_constructor_calls[0]
    assert a_base_constructor_call == b.constructor

    assert len(b.inheritance) == 1
    assert b.inheritance[0] == c
    assert len(b.immediate_inheritance) == 1
    assert b.immediate_inheritance[0] == c
    assert len(b.explicit_base_constructor_calls) == 0


def test_inheritance_with_duplicate_names(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.0")
    Slither(
        Path(TEST_DATA_DIR / "duplicate_names", "contract_with_duplicate_names.sol").as_posix(),
        solc=solc_path,
    )



from slither.utils.output import Output
from slither.core.compilation_unit import SlitherCompilationUnit
from slither.formatters.variables.unchanged_state_variables import custom_format
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from .unchanged_state_variables import UnchangedStateVariables


class CouldBeConstant(AbstractDetector):
    """
    State variables that could be declared as constant.
    Not all types for constants are implemented in Solidity as of 0.4.25.
    The only supported types are value types and strings (ElementaryType).
    Reference: https://solidity.readthedocs.io/en/latest/contracts.html#constant-state-variables
    """

    ARGUMENT = "constable-states"
    HELP = "State variables that could be declared constant"
    IMPACT = DetectorClassification.OPTIMIZATION
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#state-variables-that-could-be-declared-constant"

    WIKI_TITLE = "State variables that could be declared constant"
    WIKI_DESCRIPTION = "State variables that are not updated following deployment should be declared constant to save gas."
    WIKI_RECOMMENDATION = "Add the `constant` attribute to state variables that never change."

    def _detect(self) -> list[Output]:
        """Detect state variables that could be constant"""
        results = {}

        unchanged_state_variables = UnchangedStateVariables(self.compilation_unit)
        unchanged_state_variables.detect()

        for variable in unchanged_state_variables.constant_candidates:
            results[variable.canonical_name] = self.generate_result(
                [variable, " should be constant \n"]
            )

        # Order by canonical name for deterministic results
        return [results[k] for k in sorted(results)]

    @staticmethod
    def _format(compilation_unit: SlitherCompilationUnit, result: dict) -> None:
        custom_format(compilation_unit, result, "constant")



from slither.slithir.operations import Binary, BinaryType
from slither.tools.mutator.utils.patch import create_patch_with_line
from slither.tools.mutator.mutators.abstract_mutator import AbstractMutator

bitwise_operators = [
    BinaryType.AND,
    BinaryType.OR,
    BinaryType.LEFT_SHIFT,
    BinaryType.RIGHT_SHIFT,
    BinaryType.CARET,
]


class BOR(AbstractMutator):
    NAME = "BOR"
    HELP = "Bitwise Operator Replacement"

    def _mutate(self) -> dict:
        result: dict = {}

        for function in self.contract.functions_and_modifiers_declared:
            if not self.should_mutate_function(function):
                continue
            for node in function.nodes:
                if not self.should_mutate_node(node):
                    continue
                for ir in node.irs:
                    if isinstance(ir, Binary) and ir.type in bitwise_operators:
                        alternative_ops = bitwise_operators[:]
                        alternative_ops.remove(ir.type)
                        for op in alternative_ops:
                            # Get the string
                            start = node.source_mapping.start
                            stop = start + node.source_mapping.length
                            old_str = node.source_mapping.content
                            line_no = node.source_mapping.lines
                            # Replace the expression with true
                            halves = old_str.split(ir.type.value)
                            if len(halves) != 2:
                                continue  # skip if assembly
                            new_str = f"{halves[0]}{op.value}{halves[1]}"
                            create_patch_with_line(
                                result,
                                self.in_file,
                                start,
                                stop,
                                old_str,
                                new_str,
                                line_no[0],
                            )
        return result



"""
Module detecting missing events for critical contract parameters set by owners and used in access control

"""

from slither.analyses.data_dependency.data_dependency import is_tainted
from slither.core.cfg.node import Node
from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.core.declarations.modifier import Modifier
from slither.core.solidity_types.elementary_type import ElementaryType
from slither.core.variables.state_variable import StateVariable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations.event_call import EventCall
from slither.utils.output import Output


class MissingEventsAccessControl(AbstractDetector):
    """
    Missing events for critical contract parameters set by owners and used in access control
    """

    ARGUMENT = "events-access"
    HELP = "Missing Events Access Control"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#missing-events-access-control"
    WIKI_TITLE = "Missing events access control"
    WIKI_DESCRIPTION = "Detect missing events for critical access control parameters"

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract C {

  modifier onlyAdmin {
    if (msg.sender != owner) throw;
    _;
  }

  function updateOwner(address newOwner) onlyAdmin external {
    owner = newOwner;
  }
}
```
`updateOwner()` has no event, so it is difficult to track off-chain owner changes.
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Emit an event for critical parameter changes."

    @staticmethod
    def _detect_missing_events(
        contract: Contract,
    ) -> list[tuple[FunctionContract, list[tuple[Node, StateVariable, Modifier]]]]:
        """
        Detects if critical contract parameters set by owners and used in access control are missing events
        :param contract: The contract to check
        :return: Functions with nodes of critical operations but no events
        """
        results = []

        for function in contract.functions_entry_points:
            nodes = []

            # Check for any events in the function and skip if found
            # Note: not checking if event corresponds to critical parameter
            if any(ir for node in function.nodes for ir in node.irs if isinstance(ir, EventCall)):
                continue

            # Ignore constructors and private/internal functions
            # Heuristic-1: functions with critical operations are typically "protected". Skip unprotected functions.
            if function.is_constructor or not function.is_protected():
                continue

            # Heuristic-2: Critical operations are where state variables are written and tainted
            # Heuristic-3: Variables of interest are address type that are used in modifiers i.e. access control
            # Heuristic-4: Critical operations present but no events in the function is not a good practice
            for node in function.nodes:
                for sv in node.state_variables_written:
                    if is_tainted(sv, function) and sv.type == ElementaryType("address"):
                        for mod in function.contract.modifiers:
                            if sv in mod.state_variables_read:
                                nodes.append((node, sv, mod))
            if nodes:
                results.append((function, nodes))
        return results

    def _detect(self) -> list[Output]:
        """Detect missing events for critical contract parameters set by owners and used in access control
        Returns:
            list: {'(function, node)'}
        """

        # Check derived contracts for missing events
        results = []
        for contract in self.compilation_unit.contracts_derived:
            missing_events = self._detect_missing_events(contract)
            for function, nodes in missing_events:
                info: DETECTOR_INFO = [function, " should emit an event for: \n"]
                for node, _sv, _mod in nodes:
                    info += ["\t- ", node, " \n"]
                res = self.generate_result(info)
                results.append(res)
        return results



from slither.core.variables.variable import Variable


class EventVariable(Variable):
    def __init__(self) -> None:
        super().__init__()
        self._indexed = False

    @property
    def indexed(self) -> bool:
        """
        Indicates whether the event variable is indexed in the bloom filter.
        :return: Returns True if the variable is indexed in bloom filter, False otherwise.
        """
        return self._indexed

    @indexed.setter
    def indexed(self, is_indexed: bool) -> None:
        self._indexed = is_indexed



from slither.core.variables.local_variable import LocalVariable
from slither.vyper_parsing.ast.types import Arg, Name, AnnAssign, Subscript, Call, Tuple
from slither.vyper_parsing.type_parsing import parse_type


class LocalVariableVyper:
    def __init__(self, variable: LocalVariable, variable_data: Arg | AnnAssign | Name) -> None:
        self._variable: LocalVariable = variable

        if isinstance(variable_data, Arg):
            self._variable.name = variable_data.arg
            self._elem_to_parse = variable_data.annotation
        elif isinstance(variable_data, AnnAssign):
            self._variable.name = variable_data.target.id
            self._elem_to_parse = variable_data.annotation
        else:
            assert isinstance(variable_data, Name)
            self._variable.name = variable_data.id
            self._elem_to_parse = variable_data

        assert isinstance(self._elem_to_parse, (Name, Subscript, Call, Tuple))

        # Vyper does not have data locations or storage pointers.
        # If this was left as default, reference types would be considered storage by `LocalVariable.is_storage`
        self._variable.set_location("memory")

    @property
    def underlying_variable(self) -> LocalVariable:
        return self._variable

    def analyze(self, contract) -> None:
        self._variable.type = parse_type(self._elem_to_parse, contract)



from slither.solc_parsing.variables.variable_declaration import VariableDeclarationSolc
from slither.core.variables.structure_variable import StructureVariable


class StructureVariableSolc(VariableDeclarationSolc):
    def __init__(self, variable: StructureVariable, variable_data: dict):
        super().__init__(variable, variable_data)

    @property
    def underlying_variable(self) -> StructureVariable:
        # Todo: Not sure how to overcome this with mypy
        assert isinstance(self._variable, StructureVariable)
        return self._variable



import sys
import logging
import argparse
from crytic_compile import cryticparser
from slither.tools.kspec_coverage.kspec_coverage import kspec_coverage

logging.basicConfig()
logger = logging.getLogger("Slither.kspec")
logger.setLevel(logging.INFO)

ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(message)s")
logger.addHandler(ch)
logger.handlers[0].setFormatter(formatter)
logger.propagate = False


def parse_args() -> argparse.Namespace:
    """
    Parse the underlying arguments for the program.
    :return: Returns the arguments for the program.
    """
    parser = argparse.ArgumentParser(
        description="slither-kspec-coverage",
        usage="slither-kspec-coverage contract.sol kspec.md",
    )

    parser.add_argument(
        "contract", help="The filename of the contract or truffle directory to analyze."
    )
    parser.add_argument(
        "kspec",
        help="The filename of the Klab spec markdown for the analyzed contract(s)",
    )

    parser.add_argument(
        "--version",
        help="displays the current version",
        version="0.1.0",
        action="version",
    )
    parser.add_argument(
        "--json",
        help='Export the results as a JSON file ("--json -" to export to stdout)',
        action="store",
        default=False,
    )

    cryticparser.init(parser)

    if len(sys.argv) < 2:
        parser.print_help(sys.stderr)
        sys.exit(1)

    return parser.parse_args()


def main() -> None:
    # ------------------------------
    #       Usage: slither-kspec-coverage contract kspec
    #       Example: slither-kspec-coverage contract.sol kspec.md
    # ------------------------------
    # Parse all arguments

    args = parse_args()

    kspec_coverage(args)


if __name__ == "__main__":
    main()



"""
Module detecting unused transfer/transferFrom return values from external calls
"""

from slither.core.declarations import Function
from slither.detectors.abstract_detector import DetectorClassification
from slither.detectors.operations.unused_return_values import UnusedReturnValues
from slither.slithir.operations import HighLevelCall
from slither.slithir.operations.operation import Operation


class UncheckedTransfer(UnusedReturnValues):
    """
    If the return value of a transfer/transferFrom function is never used, it's likely to be bug
    """

    ARGUMENT = "unchecked-transfer"
    HELP = "Unchecked tokens transfer"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#unchecked-transfer"

    WIKI_TITLE = "Unchecked transfer"
    WIKI_DESCRIPTION = "The return value of an external transfer/transferFrom call is not checked"

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Token {
    function transferFrom(address _from, address _to, uint256 _value) public returns (bool success);
}
contract MyBank{
    mapping(address => uint) balances;
    Token token;
    function deposit(uint amount) public{
        token.transferFrom(msg.sender, address(this), amount);
        balances[msg.sender] += amount;
    }
}
```
Several tokens do not revert in case of failure and return false. If one of these tokens is used in `MyBank`, `deposit` will not revert if the transfer fails, and an attacker can call `deposit` for free.."""
    # endregion wiki_exploit_scenariox

    WIKI_RECOMMENDATION = (
        "Use `SafeERC20`, or ensure that the transfer/transferFrom return value is checked."
    )

    def _is_instance(self, ir: Operation) -> bool:
        return (
            isinstance(ir, HighLevelCall)
            and isinstance(ir.function, Function)
            and ir.function.solidity_signature
            in ["transfer(address,uint256)", "transferFrom(address,address,uint256)"]
        )



from contextlib import contextmanager
import logging
from slither.utils.colors import bold, yellow, red


@contextmanager
def snip_section(message: str | None) -> None:
    if message:
        print(red(message), end="\n\n")

    print(yellow("---- snip 8< ----"))
    yield
    print(yellow("---- >8 snip ----"))


@contextmanager
def report_section(title: str) -> None:
    print(bold(f"## {title}"), end="\n\n")
    try:
        yield
    except Exception as e:
        with snip_section(
            "slither-doctor failed unexpectedly! Please report this on the Slither GitHub issue tracker, and include the output below:"
        ):
            logging.exception(e)
    finally:
        print(end="\n\n")



# Function computing the code complexity
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from slither.core.declarations import Function
    from slither.core.cfg.node import Node


def compute_number_edges(function: "Function") -> int:
    """
    Compute the number of edges of the CFG
    Args:
        function (core.declarations.function.Function)
    Returns:
        int
    """
    n = 0
    for node in function.nodes:
        n += len(node.sons)
    return n


def compute_strongly_connected_components(function: "Function") -> list[list["Node"]]:
    """
        Compute strongly connected components
        Based on Kosaraju algo
        Implem follows wikipedia algo: https://en.wikipedia.org/wiki/Kosaraju%27s_algorithm#The_algorithm
    Args:
        function (core.declarations.function.Function)
    Returns:
        list(list(nodes))
    """
    visited = dict.fromkeys(function.nodes, False)
    assigned = dict.fromkeys(function.nodes, False)
    components = []
    l = []

    def visit(node: "Node") -> None:
        if not visited[node]:
            visited[node] = True
            for son in node.sons:
                visit(son)
            l.append(node)

    for n in function.nodes:
        visit(n)

    def assign(node: "Node", root: list["Node"]) -> None:
        if not assigned[node]:
            assigned[node] = True
            root.append(node)
            for father in node.fathers:
                assign(father, root)

    for n in reversed(l):
        component: list[Node] = []
        assign(n, component)
        if component:
            components.append(component)

    return components


def compute_cyclomatic_complexity(function: "Function") -> int:
    """
    Compute the cyclomatic complexity of a function
    Args:
        function (core.declarations.function.Function)
    Returns:
        int
    """
    # from https://en.wikipedia.org/wiki/Cyclomatic_complexity
    # M = E - N + 2P
    # where M is the complexity
    # E number of edges
    # N number of nodes
    # P number of connected components (always 1 for a function)

    E = compute_number_edges(function)
    N = len(function.nodes)
    P = 1
    return E - N + 2 * P



"""
Module printing the inheritance graph

The inheritance graph shows the relation between the contracts
and their functions/modifiers/public variables.
The output is a dot file named filename.dot
"""

from slither.core.declarations.contract import Contract
from slither.core.solidity_types.user_defined_type import UserDefinedType
from slither.printers.abstract_printer import AbstractPrinter
from slither.utils.inheritance_analysis import (
    detect_c3_function_shadowing,
    detect_state_variable_shadowing,
)


def _get_pattern_func(func):
    # Html pattern, each line is a row in a table
    func_name = func.full_name
    pattern = '<TR><TD align="left">    %s</TD></TR>'
    pattern_shadow = '<TR><TD align="left"><font color="#FFA500">    %s</font></TD></TR>'
    if func.shadows:
        return pattern_shadow % func_name
    return pattern % func_name


def _get_port_id(var, contract):
    return f"{var.name}{contract.name}"


class PrinterInheritanceGraph(AbstractPrinter):
    ARGUMENT = "inheritance-graph"
    HELP = "Export the inheritance graph of each contract to a dot file"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#inheritance-graph"

    def __init__(self, slither, logger):
        super().__init__(slither, logger)

        inheritance = [x.inheritance for x in slither.contracts]
        self.inheritance = {item for sublist in inheritance for item in sublist}

        self.overshadowing_state_variables = {}
        shadows = detect_state_variable_shadowing(slither.contracts)
        for overshadowing_instance in shadows:
            overshadowing_state_var = overshadowing_instance[1]
            overshadowed_state_var = overshadowing_instance[3]

            # Add overshadowing variable entry.
            if overshadowing_state_var not in self.overshadowing_state_variables:
                self.overshadowing_state_variables[overshadowing_state_var] = set()
            self.overshadowing_state_variables[overshadowing_state_var].add(overshadowed_state_var)

    def _get_pattern_var(self, var):
        # Html pattern, each line is a row in a table
        var_name = var.name
        pattern = '<TR><TD align="left">    %s</TD></TR>'
        pattern_contract = (
            '<TR><TD align="left">    %s<font color="blue" POINT-SIZE="10"> (%s)</font></TD></TR>'
        )
        pattern_shadow = '<TR><TD align="left"><font color="red">    %s</font></TD></TR>'
        pattern_contract_shadow = '<TR><TD align="left"><font color="red">    %s</font><font color="blue" POINT-SIZE="10"> (%s)</font></TD></TR>'

        if isinstance(var.type, UserDefinedType) and isinstance(var.type.type, Contract):
            if var in self.overshadowing_state_variables:
                return pattern_contract_shadow % (var_name, var.type.type.name)
            return pattern_contract % (var_name, var.type.type.name)

        if var in self.overshadowing_state_variables:
            return pattern_shadow % var_name
        return pattern % var_name

    @staticmethod
    def _get_indirect_shadowing_information(contract):
        """
        Obtain a string that describes variable shadowing for the given variable. None if no shadowing exists.
        :param var: The variable to collect shadowing information for.
        :param contract: The contract in which this variable is being analyzed.
        :return: Returns a string describing variable shadowing for the given variable. None if no shadowing exists.
        """
        # If this variable is an overshadowing variable, we'll want to return information describing it.
        result = []
        indirect_shadows = detect_c3_function_shadowing(contract)
        for winner, colliding_functions in indirect_shadows.items():
            collision_steps = ", ".join(
                [f.contract_declarer.name for f in colliding_functions]
                + [winner.contract_declarer.name]
            )
            result.append(
                f"'{winner.full_name}' collides in inherited contracts {collision_steps} where {winner.contract_declarer.name} is chosen."
            )
        return "\n".join(result)

    def _summary(self, contract):
        """
        Build summary using HTML
        """
        ret = ""

        # Remove contracts that have "mock" in the name and if --include-interfaces in False (default)
        # removes inherited interfaces
        inheritance = [
            i
            for i in contract.immediate_inheritance
            if "mock" not in i.name.lower()
            and (not i.is_interface or self.slither.include_interfaces)
        ]

        # Add arrows (number them if there is more than one path so we know order of declaration for inheritance).
        if len(inheritance) == 1:
            immediate_inheritance = contract.immediate_inheritance[0]
            ret += f"c{contract.id}_{contract.name} -> c{immediate_inheritance.id}_{immediate_inheritance};\n"
        else:
            for i, immediate_inheritance in enumerate(inheritance):
                ret += f'c{contract.id}_{contract.name} -> c{immediate_inheritance.id}_{immediate_inheritance} [ label="{i + 1}" ];\n'

        # Functions
        visibilities = ["public", "external"]
        public_functions = [
            _get_pattern_func(f)
            for f in contract.functions
            if not f.is_constructor
            and not f.is_constructor_variables
            and not f.is_virtual
            and f.contract_declarer == contract
            and f.visibility in visibilities
        ]
        public_functions = "".join(public_functions)
        private_functions = [
            _get_pattern_func(f)
            for f in contract.functions
            if not f.is_constructor
            and not f.is_constructor_variables
            and f.contract_declarer == contract
            and f.visibility not in visibilities
        ]
        private_functions = "".join(private_functions)

        # Modifiers
        modifiers = [
            _get_pattern_func(m) for m in contract.modifiers if m.contract_declarer == contract
        ]
        modifiers = "".join(modifiers)

        # Public variables
        public_variables = [
            self._get_pattern_var(v)
            for v in contract.state_variables_declared
            if v.visibility in visibilities
        ]
        public_variables = "".join(public_variables)

        private_variables = [
            self._get_pattern_var(v)
            for v in contract.state_variables_declared
            if v.visibility not in visibilities
        ]
        private_variables = "".join(private_variables)

        # Obtain any indirect shadowing information for this node.
        indirect_shadowing_information = self._get_indirect_shadowing_information(contract)

        # Build the node label
        ret += f'c{contract.id}_{contract.name}[shape="box"'
        ret += 'label=< <TABLE border="0">'
        ret += f'<TR><TD align="center"><B>{contract.name}</B></TD></TR>'
        if public_functions:
            ret += '<TR><TD align="left"><I>Public Functions:</I></TD></TR>'
            ret += f"{public_functions}"
        if private_functions:
            ret += '<TR><TD align="left"><I>Private Functions:</I></TD></TR>'
            ret += f"{private_functions}"
        if modifiers:
            ret += '<TR><TD align="left"><I>Modifiers:</I></TD></TR>'
            ret += f"{modifiers}"
        if public_variables:
            ret += '<TR><TD align="left"><I>Public Variables:</I></TD></TR>'
            ret += f"{public_variables}"
        if private_variables:
            ret += '<TR><TD align="left"><I>Private Variables:</I></TD></TR>'
            ret += f"{private_variables}"

        if indirect_shadowing_information:
            ret += f'<TR><TD><BR/></TD></TR><TR><TD align="left" border="1"><font color="#777777" point-size="10">{indirect_shadowing_information.replace(chr(10), "<BR/>")}</font></TD></TR>'
        ret += "</TABLE> >];\n"

        return ret

    def output(self, filename):
        """
        Output the graph in filename
        Args:
            filename(string)
        """
        if filename in ("", "."):
            filename = "inheritance-graph.dot"
        if not filename.endswith(".dot"):
            filename += ".inheritance-graph.dot"
        info = "Inheritance Graph: " + filename + "\n"
        self.info(info)

        content = 'digraph "" {\n'
        for c in self.contracts:
            if (
                "mock" in c.name.lower()
                or c.is_library
                or (c.is_interface and not self.slither.include_interfaces)
            ):
                continue
            content += self._summary(c) + "\n"
        content += "}"

        with open(filename, "w", encoding="utf8") as f:
            f.write(content)

        res = self.generate_output(info)
        res.add_file(filename, content)

        return res



"""
Testing utilities for the read-storage tool
"""

import shutil
import subprocess
from time import sleep
from collections.abc import Generator
from dataclasses import dataclass
from web3 import Web3
import pytest


@dataclass
class GanacheInstance:
    def __init__(self, provider: str, eth_address: str, eth_privkey: str):
        self.provider = provider
        self.eth_address = eth_address
        self.eth_privkey = eth_privkey


@pytest.fixture(scope="module", name="ganache")
def fixture_ganache() -> Generator[GanacheInstance, None, None]:
    """Fixture that runs ganache"""
    if not shutil.which("ganache"):
        raise Exception(
            "ganache was not found in PATH, you can install it with `npm install -g ganache`"
        )

    # Address #1 when ganache is run with `--wallet.seed test`, it starts with 1000 ETH
    eth_address = "0xae17D2dD99e07CA3bF2571CCAcEAA9e2Aefc2Dc6"
    eth_privkey = (
        "0xe48ba530a63326818e116be262fd39ae6dcddd89da4b1f578be8afd4e8894b8d"  # gitleaks:allow
    )
    eth = int(1e18 * 1e6)
    port = 8545
    with subprocess.Popen(
        f"""ganache
        --port {port}
        --chain.networkId 1
        --chain.chainId 1
        --account {eth_privkey},{eth}
        """.replace("\n", " "),
        shell=True,
    ) as p:
        sleep(3)
        yield GanacheInstance(f"http://127.0.0.1:{port}", eth_address, eth_privkey)
        p.kill()
        p.wait()


@pytest.fixture(scope="module", name="web3")
def fixture_web3(ganache: GanacheInstance):
    w3 = Web3(Web3.HTTPProvider(ganache.provider, request_kwargs={"timeout": 30}))
    return w3



from pathlib import Path
from slither import Slither
from slither.slithir.operations import Assignment
from slither.slithir.variables import Constant

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"


func_to_results = {
    "returnEvent()": "16700440330922901039223184000601971290390760458944929668086539975128325467771",
    "returnError()": "224292994",
    "returnFunctionFromContract()": "890000139",
    "returnFunction()": "890000139",
    "returnFunctionWithStructure()": "1430834845",
    "returnFunctionThroughLocaLVar()": "3781905051",
}


def test_enum_max_min(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.19")
    slither = Slither(Path(TEST_DATA_DIR, "selector.sol").as_posix(), solc=solc_path)

    contract = slither.get_contract_from_name("Test")[0]

    for func_name, value in func_to_results.items():
        f = contract.get_function_from_signature(func_name)
        assignment = f.slithir_operations[0]
        assert (
            isinstance(assignment, Assignment)
            and isinstance(assignment.rvalue, Constant)
            and assignment.rvalue.value == value
        )



import abc

from slither.core.source_mapping.source_mapping import SourceMapping


class Type(SourceMapping, metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def storage_size(self) -> tuple[int, bool]:
        """
        Computes and returns storage layout related metadata

        :return: (int, bool) - the number of bytes this type will require, and whether it must start in
        a new slot regardless of whether the current slot can still fit it
        """

    @property
    @abc.abstractmethod
    def is_dynamic(self) -> bool:
        """True if the size of the type is dynamic"""



from slither.analyses.data_dependency.data_dependency import is_dependent
from slither.core.cfg.node import Node
from slither.core.compilation_unit import SlitherCompilationUnit
from slither.core.declarations import Contract, Function, SolidityVariableComposed, FunctionContract
from slither.core.declarations.solidity_variables import SolidityVariable
from slither.slithir.operations import HighLevelCall, LibraryCall


class ArbitrarySendErc20:
    """Detects instances where ERC20 can be sent from an arbitrary from address."""

    def __init__(self, compilation_unit: SlitherCompilationUnit) -> None:
        self._compilation_unit = compilation_unit
        self._no_permit_results: list[Node] = []
        self._permit_results: list[Node] = []

    @property
    def compilation_unit(self) -> SlitherCompilationUnit:
        return self._compilation_unit

    @property
    def no_permit_results(self) -> list[Node]:
        return self._no_permit_results

    @property
    def permit_results(self) -> list[Node]:
        return self._permit_results

    def _detect_arbitrary_from(self, contract: Contract) -> None:
        for f in contract.functions:
            all_high_level_calls = [
                ir.function.solidity_signature
                for _, ir in f.high_level_calls
                if isinstance(ir.function, Function)
            ]
            all_library_calls = [ir.function.solidity_signature for ir in f.library_calls]
            if (
                "transferFrom(address,address,uint256)" in all_high_level_calls
                or "safeTransferFrom(address,address,address,uint256)" in all_library_calls
            ):
                if (
                    "permit(address,address,uint256,uint256,uint8,bytes32,bytes32)"
                    in all_high_level_calls
                ):
                    ArbitrarySendErc20._arbitrary_from(f, self._permit_results)
                else:
                    ArbitrarySendErc20._arbitrary_from(f, self._no_permit_results)

    @staticmethod
    def _arbitrary_from(function: FunctionContract, results: list[Node]) -> None:
        """Finds instances of (safe)transferFrom that do not use msg.sender or address(this) as from parameter."""
        for _, ir in function.high_level_calls:
            if (
                isinstance(ir, LibraryCall)
                and ir.function.solidity_signature
                == "safeTransferFrom(address,address,address,uint256)"
                and not (
                    is_dependent(
                        ir.arguments[1],
                        SolidityVariableComposed("msg.sender"),
                        ir.node,
                    )
                    or is_dependent(
                        ir.arguments[1],
                        SolidityVariable("this"),
                        ir.node,
                    )
                )
            ):
                results.append(ir.node)
            elif (
                isinstance(ir, HighLevelCall)
                and isinstance(ir.function, Function)
                and ir.function.solidity_signature == "transferFrom(address,address,uint256)"
                and not (
                    is_dependent(
                        ir.arguments[0],
                        SolidityVariableComposed("msg.sender"),
                        ir.node,
                    )
                    or is_dependent(
                        ir.arguments[0],
                        SolidityVariable("this"),
                        ir.node,
                    )
                )
            ):
                results.append(ir.node)

    def detect(self) -> None:
        """Detect transfers that use arbitrary `from` parameter."""
        for c in self.compilation_unit.contracts_derived:
            self._detect_arbitrary_from(c)



"""
Cheatcode printer

This printer prints the usage of cheatcode in the code.
"""

from slither.printers.abstract_printer import AbstractPrinter
from slither.slithir.operations import HighLevelCall
from slither.utils import output


class CheatcodePrinter(AbstractPrinter):
    ARGUMENT = "cheatcode"

    HELP = """
        Print the usage of (Foundry) cheatcodes in the code.
        For the complete list of Cheatcodes, see https://book.getfoundry.sh/cheatcodes/
    """

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#cheatcode"

    def output(self, filename: str) -> output.Output:
        info: str = ""

        try:
            vm = self.slither.get_contract_from_name("Vm").pop()
        except IndexError:
            return output.Output("No contract named VM found")

        for contract in self.slither.contracts_derived:
            # Check that the IS_TEST variable is set. (Only works for Foundry)
            is_test_var = contract.variables_as_dict.get("IS_TEST", None)
            is_test = False
            if is_test_var is not None:
                try:
                    is_test = is_test_var.expression.value == "true"
                except AttributeError:
                    pass

            if not is_test:
                continue

            found_contract: bool = False
            contract_info: str = ""
            for func in contract.functions_declared:
                function_info = f"\t{func}\n"
                found_function: bool = False
                for node in func.nodes:
                    for op in node.all_slithir_operations():
                        if (
                            isinstance(op, HighLevelCall)
                            and op.function.contract == vm
                            and op.function.visibility == "external"
                        ):
                            found_function = True
                            function_info += (
                                f"\t\t{op.function.name} - ({node.source_mapping.to_detailed_str()})\n"
                                f"\t\t{node.expression}\n\n"
                            )

                if found_function:
                    if found_contract is False:
                        contract_info = f"{contract} ({contract.source_mapping.filename.short})\n"
                        found_contract = True

                    contract_info += function_info

            if found_contract:
                info += contract_info

        self.info(info)
        res = output.Output(info)
        return res



import logging

from slither.core.declarations import Function
from slither.slithir.operations import InternalCall
from slither.tools.upgradeability.checks.abstract_checks import (
    AbstractCheck,
    CheckClassification,
)
from slither.utils.colors import red
from slither.exceptions import SlitherError

logger = logging.getLogger("Slither-check-upgradeability")


class MultipleInitTarget(Exception):
    pass


def _has_initialize_modifier(function: Function):
    if not function.modifiers:
        return False
    return any((m.name in ("initializer", "reinitializer")) for m in function.modifiers)


def _get_initialize_functions(contract):
    return [
        f
        for f in contract.functions
        if (f.name == "initialize" or _has_initialize_modifier(f)) and f.is_implemented
    ]


def _get_all_internal_calls(function):
    all_ir = function.all_slithir_operations()
    return [
        i.function
        for i in all_ir
        if isinstance(i, InternalCall) and i.function_name == "initialize"
    ]


def _get_most_derived_init(contract):
    init_functions = [f for f in contract.functions if not f.is_shadowed and f.name == "initialize"]
    if len(init_functions) > 1:
        if len([f for f in init_functions if f.contract_declarer == contract]) == 1:
            return next(f for f in init_functions if f.contract_declarer == contract)
        raise MultipleInitTarget
    if init_functions:
        return init_functions[0]
    return None


class InitializablePresent(AbstractCheck):
    ARGUMENT = "init-missing"
    IMPACT = CheckClassification.INFORMATIONAL

    HELP = "Initializable is missing"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#initializable-is-missing"
    WIKI_TITLE = "Initializable is missing"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect if a contract `Initializable` is present.
"""
    # endregion wiki_description

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Review manually the contract's initialization..
Consider using a `Initializable` contract to follow [standard practice](https://docs.openzeppelin.com/upgrades/2.7/writing-upgradeable).
"""
    # endregion wiki_recommendation

    def _check(self):
        initializable = self.contract.file_scope.get_contract_from_name("Initializable")
        if initializable is None:
            info = [
                "Initializable contract not found, the contract does not follow a standard initalization schema.\n"
            ]
            json = self.generate_result(info)
            return [json]
        return []


class InitializableInherited(AbstractCheck):
    ARGUMENT = "init-inherited"
    IMPACT = CheckClassification.INFORMATIONAL

    HELP = "Initializable is not inherited"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#initializable-is-not-inherited"
    WIKI_TITLE = "Initializable is not inherited"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect if `Initializable` is inherited.
"""
    # endregion wiki_description

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Review manually the contract's initialization. Consider inheriting `Initializable`.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True

    def _check(self):
        initializable = self.contract.file_scope.get_contract_from_name("Initializable")
        # See InitializablePresent
        if initializable is None:
            return []
        if initializable not in self.contract.inheritance:
            info = [self.contract, " does not inherit from ", initializable, ".\n"]
            json = self.generate_result(info)
            return [json]
        return []


class InitializableInitializer(AbstractCheck):
    ARGUMENT = "initializer-missing"
    IMPACT = CheckClassification.INFORMATIONAL

    HELP = "initializer() is missing"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#initializer-is-missing"
    WIKI_TITLE = "initializer() is missing"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect the lack of `Initializable.initializer()` modifier.
"""
    # endregion wiki_description

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Review manually the contract's initialization. Consider inheriting a `Initializable.initializer()` modifier.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True

    def _check(self):
        initializable = self.contract.file_scope.get_contract_from_name("Initializable")
        # See InitializablePresent
        if initializable is None:
            return []
        # See InitializableInherited
        if initializable not in self.contract.inheritance:
            return []

        initializer = self.contract.get_modifier_from_canonical_name("Initializable.initializer()")
        if initializer is None:
            info = ["Initializable.initializer() does not exist.\n"]
            json = self.generate_result(info)
            return [json]
        return []


class MissingInitializerModifier(AbstractCheck):
    ARGUMENT = "missing-init-modifier"
    IMPACT = CheckClassification.HIGH

    HELP = "initializer() is not called"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#initializer-is-not-called"
    WIKI_TITLE = "initializer() is not called"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect if `Initializable.initializer()` or `Initializable.reinitializer(uint64)` is called.
"""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Contract{
    function initialize() public{
        ///
    }
}

```
`initialize` should have the `initializer` modifier to prevent someone from initializing the contract multiple times.
"""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Use `Initializable.initializer()` or `Initializable.reinitializer(uint64)`.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True

    def _check(self):
        initializable = self.contract.file_scope.get_contract_from_name("Initializable")
        # See InitializablePresent
        if initializable is None:
            return []
        # See InitializableInherited
        if initializable not in self.contract.inheritance:
            return []
        initializer = self.contract.get_modifier_from_canonical_name("Initializable.initializer()")
        reinitializer = self.contract.get_modifier_from_canonical_name(
            "Initializable.reinitializer(uint64)"
        )
        # InitializableInitializer
        if initializer is None and reinitializer is None:
            return []

        results = []
        all_init_functions = _get_initialize_functions(self.contract)
        for f in all_init_functions:
            if initializer not in f.modifiers and reinitializer not in f.modifiers:
                info = [f, " does not call the initializer or reinitializer modifier.\n"]
                json = self.generate_result(info)
                results.append(json)
        return results


class MissingCalls(AbstractCheck):
    ARGUMENT = "missing-calls"
    IMPACT = CheckClassification.HIGH

    HELP = "Missing calls to init functions"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#initialize-functions-are-not-called"
    WIKI_TITLE = "Initialize functions are not called"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect missing calls to initialize functions.
"""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Base{
    function initialize() public{
        ///
    }
}
contract Derived is Base{
    function initialize() public{
        ///
    }
}

```
`Derived.initialize` does not call `Base.initialize` leading the contract to not be correctly initialized.
"""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Ensure all the initialize functions are reached by the most derived initialize function.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True

    def _check(self):
        results = []

        # TODO: handle MultipleInitTarget
        try:
            most_derived_init = _get_most_derived_init(self.contract)
        except MultipleInitTarget:
            logger.error(red(f"Too many init targets in {self.contract}"))
            return []

        if most_derived_init is None:
            return []

        all_init_functions = _get_initialize_functions(self.contract)
        all_init_functions_called = _get_all_internal_calls(most_derived_init) + [most_derived_init]
        missing_calls = [f for f in all_init_functions if f not in all_init_functions_called]
        for f in missing_calls:
            # we don't account reinitializers
            if any((m.name == "reinitializer") for m in f.modifiers):
                continue
            info = ["Missing call to ", f, " in ", most_derived_init, ".\n"]
            json = self.generate_result(info)
            results.append(json)
        return results


class MultipleCalls(AbstractCheck):
    ARGUMENT = "multiple-calls"
    IMPACT = CheckClassification.HIGH

    HELP = "Init functions called multiple times"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#initialize-functions-are-called-multiple-times"
    WIKI_TITLE = "Initialize functions are called multiple times"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect multiple calls to a initialize function.
"""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Base{
    function initialize(uint) public{
        ///
    }
}
contract Derived is Base{
    function initialize(uint a, uint b) public{
        initialize(a);
    }
}

contract DerivedDerived is Derived{
    function initialize() public{
        initialize(0);
        initialize(0, 1 );
    }
}

```
`Base.initialize(uint)` is called two times in `DerivedDerived.initialize` execution, leading to a potential corruption.
"""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Call only one time every initialize function.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True

    def _check(self):
        results = []

        # TODO: handle MultipleInitTarget
        try:
            most_derived_init = _get_most_derived_init(self.contract)
        except MultipleInitTarget:
            # Should be already reported by MissingCalls
            # logger.error(red(f'Too many init targets in {self.contract}'))
            return []

        if most_derived_init is None:
            return []

        all_init_functions_called = _get_all_internal_calls(most_derived_init) + [most_derived_init]
        double_calls = list(
            {f for f in all_init_functions_called if all_init_functions_called.count(f) > 1}
        )
        for f in double_calls:
            info = [f, " is called multiple times in ", most_derived_init, ".\n"]
            json = self.generate_result(info)
            results.append(json)

        return results


class InitializeTarget(AbstractCheck):
    ARGUMENT = "initialize-target"
    IMPACT = CheckClassification.INFORMATIONAL

    HELP = "Initialize function that must be called"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#initialize-function"
    WIKI_TITLE = "Initialize function"

    # region wiki_description
    WIKI_DESCRIPTION = """
Show the function that must be called at deployment.

This finding does not have an immediate security impact and is informative.
"""
    # endregion wiki_description

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Ensure that the function is called at deployment.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True

    def _check(self):
        # TODO: handle MultipleInitTarget
        try:
            most_derived_init = _get_most_derived_init(self.contract)
        except MultipleInitTarget:
            # Should be already reported by MissingCalls
            # logger.error(red(f'Too many init targets in {self.contract}'))
            return []

        if most_derived_init is None:
            return []

        info = [
            self.contract,
            " needs to be initialized by ",
            most_derived_init,
            ".\n",
        ]
        json = self.generate_result(info)
        return [json]


class MultipleReinitializers(AbstractCheck):
    ARGUMENT = "multiple-new-reinitializers"
    IMPACT = CheckClassification.LOW

    HELP = "Multiple new reinitializers in the updated contract"
    WIKI = (
        "https://github.com/crytic/slither/wiki/Upgradeability-Checks#multiple-new-reinitializers"
    )
    WIKI_TITLE = "Multiple new reinitializers in the updated contract"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect multiple new reinitializers in the updated contract`.
"""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Counter is Initializable {
    uint256 public x;

    function initialize(uint256 _x) public initializer {
        x = _x;
    }

    function changeX() public {
        x++;
    }
}

contract CounterV2 is Initializable {
    uint256 public x;
    uint256 public y;
    uint256 public z;

    function initializeV2(uint256 _y) public reinitializer(2) {
        y = _y;
    }

    function initializeV3(uint256 _z) public reinitializer(3) {
        z = _z;
    }

    function changeX() public {
        x = x + y + z;
    }
}
```
`CounterV2` has two reinitializers with new versions `2` and `3`. If `initializeV3()` is called first, the `initializeV2()` can't be called to initialize `y`, which may brings security risks.
"""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Do not use multiple reinitializers with higher versions in the updated contract. Please consider combining new reinitializers into a single one.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True
    REQUIRE_CONTRACT_V2 = True

    def _check(self):
        contract_v1 = self.contract
        contract_v2 = self.contract_v2

        if contract_v2 is None:
            raise SlitherError("multiple-new-reinitializers requires a V2 contract")

        initializerV1 = contract_v1.get_modifier_from_canonical_name("Initializable.initializer()")
        reinitializerV1 = contract_v1.get_modifier_from_canonical_name(
            "Initializable.reinitializer(uint64)"
        )
        reinitializerV2 = contract_v2.get_modifier_from_canonical_name(
            "Initializable.reinitializer(uint64)"
        )

        # contractV1 has initializer or reinitializer
        if initializerV1 is None and reinitializerV1 is None:
            return []
        # contractV2 has reinitializer
        if reinitializerV2 is None:
            return []

        initializer_funcs_v1 = _get_initialize_functions(contract_v1)
        initializer_funcs_v2 = _get_initialize_functions(contract_v2)
        new_reinitializer_funcs = []
        for fv2 in initializer_funcs_v2:
            if not any((fv2.full_name == fv1.full_name) for fv1 in initializer_funcs_v1):
                new_reinitializer_funcs.append(fv2)

        results = []
        if len(new_reinitializer_funcs) > 1:
            for f in new_reinitializer_funcs:
                info = [
                    f,
                    " multiple new reinitializers which should be combined into one per upgrade.\n",
                ]
                json = self.generate_result(info)
                results.append(json)
        return results



from typing import Union

from slither.core.expressions.binary_operation import BinaryOperation
from slither.core.expressions.expression import Expression
from slither.core.expressions.identifier import Identifier
from slither.core.expressions.literal import Literal
from slither.core.expressions.tuple_expression import TupleExpression
from slither.core.expressions.type_conversion import TypeConversion
from slither.core.expressions.unary_operation import UnaryOperation


class ConditionalExpression(Expression):
    def __init__(
        self,
        if_expression: BinaryOperation | Identifier | Literal,
        then_expression: Union[
            "ConditionalExpression", TypeConversion, Literal, TupleExpression, Identifier
        ],
        else_expression: TupleExpression | UnaryOperation | Identifier | Literal,
    ) -> None:
        assert isinstance(if_expression, Expression)
        assert isinstance(then_expression, Expression)
        assert isinstance(else_expression, Expression)
        super().__init__()
        self._if_expression: Expression = if_expression
        self._then_expression: Expression = then_expression
        self._else_expression: Expression = else_expression

    @property
    def expressions(self) -> list[Expression]:
        return [self._if_expression, self._then_expression, self._else_expression]

    @property
    def if_expression(self) -> Expression:
        return self._if_expression

    @property
    def else_expression(self) -> Expression:
        return self._else_expression

    @property
    def then_expression(self) -> Expression:
        return self._then_expression

    def __str__(self) -> str:
        return (
            "if "
            + str(self._if_expression)
            + " then "
            + str(self._then_expression)
            + " else "
            + str(self._else_expression)
        )



import json
from pathlib import Path
import urllib.request


def retrieve_json(url):
    with urllib.request.urlopen(url) as response:
        data = response.read().decode("utf-8")
        return json.loads(data)


def organize_data(json_data):
    version_bugs = {}
    for version, info in json_data.items():
        version_bugs[version] = info["bugs"]
    return version_bugs


if __name__ == "__main__":
    bug_list_url = (
        "https://raw.githubusercontent.com/ethereum/solidity/develop/docs/bugs_by_version.json"
    )
    bug_data = retrieve_json(bug_list_url)
    bugs_by_version = organize_data(bug_data)

    with open(Path.cwd() / Path("slither/utils/buggy_versions.py"), "w", encoding="utf-8") as file:
        file.write(f"bugs_by_version = {bugs_by_version}")



from slither.slither import Slither

slither = Slither("coin.sol")
coin = slither.get_contract_from_name("Coin")[0]

# Iterate over all the contracts
for contract in slither.contracts:
    # If the contract is derived from MyContract
    if coin in contract.inheritance:
        # Get the function definition
        mint = contract.get_function_from_signature("_mint(address,uint256)")
        # If the function was not declared by coin, there is a bug !
        # Detect error only for contracts overriding the '_mint' function
        if mint.contract_declarer == contract:
            print(f"Error, {contract} overrides {mint}")



import re

from slither.core.compilation_unit import SlitherCompilationUnit
from slither.formatters.exceptions import FormatError, FormatImpossible
from slither.formatters.utils.patches import create_patch


def custom_format(compilation_unit: SlitherCompilationUnit, result, attribute: str) -> None:
    elements = result["elements"]
    for element in elements:
        # TODO: decide if this should be changed in the constant detector
        contract_name = element["type_specific_fields"]["parent"]["name"]
        scope = compilation_unit.get_scope(element["source_mapping"]["filename_absolute"])
        contract = scope.get_contract_from_name(contract_name)
        var = contract.get_state_variable_from_name(element["name"])
        if not var.expression:
            raise FormatImpossible(f"{var.name} is uninitialized and cannot become {attribute}.")

        _patch(
            compilation_unit,
            result,
            element["source_mapping"]["filename_absolute"],
            element["name"],
            f"{attribute} " + element["name"],
            element["source_mapping"]["start"],
            element["source_mapping"]["start"] + element["source_mapping"]["length"],
        )


def _patch(
    compilation_unit: SlitherCompilationUnit,
    result,
    in_file,
    match_text,
    replace_text,
    modify_loc_start,
    modify_loc_end,
):
    in_file_str = compilation_unit.core.source_code[in_file].encode("utf8")
    old_str_of_interest = in_file_str[modify_loc_start:modify_loc_end]
    # Add keyword `constant` before the variable name
    (new_str_of_interest, num_repl) = re.subn(
        match_text, replace_text, old_str_of_interest.decode("utf8"), count=1
    )
    if num_repl != 0:
        create_patch(
            result,
            in_file,
            modify_loc_start,
            modify_loc_end,
            old_str_of_interest,
            new_str_of_interest,
        )

    else:
        raise FormatError("State variable not found?!")



from eth_typing.evm import ChecksumAddress
from eth_utils import to_int, to_text, to_checksum_address
from web3 import Web3


def get_offset_value(hex_bytes: bytes, offset: int, size: int) -> bytes:
    """
    Trims slot data to only contain the target variable's.
    Args:
        hex_bytes (HexBytes): String representation of type
        offset (int): The size (in bits) of other variables that share the same slot.
        size (int): The size (in bits) of the target variable.
    Returns:
        (bytes): The target variable's trimmed data.
    """
    size = int(size / 8)
    offset = int(offset / 8)
    if offset == 0:
        value = hex_bytes[-size:]
    else:
        start = size + offset
        value = hex_bytes[-start:-offset]
    return value


def coerce_type(solidity_type: str, value: int | str | bytes) -> int | bool | str | ChecksumAddress:
    """
    Converts input to the indicated type.
    Args:
        solidity_type (str): String representation of type.
        value (bytes): The value to be converted.
    Returns:
        (Union[int, bool, str, ChecksumAddress, hex]): The type representation of the value.
    """
    if "int" in solidity_type:
        if str(value).startswith("0x"):
            return to_int(hexstr=value)
        return to_int(value)
    if "bool" in solidity_type:
        return bool(to_int(value))
    if "string" in solidity_type and isinstance(value, bytes):
        # length * 2 is stored in lower end bits
        # TODO handle bytes and strings greater than 32 bytes
        length = int(int.from_bytes(value[-2:], "big") / 2)
        return to_text(value[:length])

    if "address" in solidity_type:
        if not isinstance(value, (str, bytes)):
            raise TypeError
        return to_checksum_address(value)  # type: ignore

    if not isinstance(value, bytes):
        raise TypeError
    return value.hex()


def get_storage_data(
    web3: Web3, checksum_address: ChecksumAddress, slot: bytes, block: int | str
) -> bytes:
    """
    Retrieves the storage data from the blockchain at target address and slot.
    Args:
        web3: Web3 instance provider.
        checksum_address (ChecksumAddress): The address to query.
        slot (bytes): The slot to retrieve data from.
        block (optional int|str): The block number to retrieve data from
    Returns:
        (HexBytes): The slot's storage data.
    """
    return bytes(web3.eth.get_storage_at(checksum_address, slot, block)).rjust(
        32, bytes(1)
    )  # pad to 32 bytes



from slither.solc_parsing.variables.variable_declaration import VariableDeclarationSolc
from slither.core.variables.state_variable import StateVariable


class StateVariableSolc(VariableDeclarationSolc):
    def __init__(self, variable: StateVariable, variable_data: dict) -> None:
        super().__init__(variable, variable_data)

    @property
    def underlying_variable(self) -> StateVariable:
        # Todo: Not sure how to overcome this with mypy
        assert isinstance(self._variable, StateVariable)
        return self._variable

    def _analyze_variable_attributes(self, attributes: dict) -> None:
        """
        Variable Location
        Can be default or transient
        """
        if "storageLocation" in attributes:
            self.underlying_variable.set_location(attributes["storageLocation"])
        else:
            # We don't have to support legacy ast
            # as transient location was added in 0.8.28
            # and we know it must be default
            self.underlying_variable.set_location("default")

        super()._analyze_variable_attributes(attributes)



from slither.core.cfg.node import Node
from slither.core.declarations import Function, FunctionContract
from slither.slithir.operations import HighLevelCall
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.utils.output import Output


class OutOfOrderRetryable(AbstractDetector):
    ARGUMENT = "out-of-order-retryable"
    HELP = "Out-of-order retryable transactions"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#out-of-order-retryable-transactions"

    WIKI_TITLE = "Out-of-order retryable transactions"
    WIKI_DESCRIPTION = "Out-of-order retryable transactions"

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract L1 {
    function doStuffOnL2() external {
        // Retryable A
        IInbox(inbox).createRetryableTicket({
            to: l2contract,
            l2CallValue: 0,
            maxSubmissionCost: maxSubmissionCost,
            excessFeeRefundAddress: msg.sender,
            callValueRefundAddress: msg.sender,
            gasLimit: gasLimit,
            maxFeePerGas: maxFeePerGas,
            data: abi.encodeCall(l2contract.claim_rewards, ())
        });
        // Retryable B
        IInbox(inbox).createRetryableTicket({
            to: l2contract,
            l2CallValue: 0,
            maxSubmissionCost: maxSubmissionCost,
            excessFeeRefundAddress: msg.sender,
            callValueRefundAddress: msg.sender,
            gasLimit: gas,
            maxFeePerGas: maxFeePerGas,
            data: abi.encodeCall(l2contract.unstake, ())
        });
    }
}

contract L2 {
    function claim_rewards() public {
        // rewards is computed based on balance and staking period
        uint unclaimed_rewards = _compute_and_update_rewards();
        token.safeTransfer(msg.sender, unclaimed_rewards);
    }

    // Call claim_rewards before unstaking, otherwise you lose your rewards
    function unstake() public {
        _free_rewards(); // clean up rewards related variables
        balance = balance[msg.sender];
        balance[msg.sender] = 0;
        staked_token.safeTransfer(msg.sender, balance);
    }
}
```
Bob calls `doStuffOnL2` but the first retryable ticket calling `claim_rewards` fails. The second retryable ticket calling `unstake` is executed successfully. As a result, Bob loses his rewards."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Do not rely on the order or successful execution of retryable tickets."

    key = "OUTOFORDERRETRYABLE"

    def _detect_multiple_tickets(
        self, function: FunctionContract, node: Node, visited: list[Node]
    ) -> None:
        if node in visited:
            return

        visited = visited + [node]

        fathers_context = []

        for father in node.fathers:
            if self.key in father.context:
                fathers_context += father.context[self.key]

        # Exclude path that dont bring further information
        if node in self.visited_all_paths:
            if all(f_c in self.visited_all_paths[node] for f_c in fathers_context):
                return
        else:
            self.visited_all_paths[node] = []

        self.visited_all_paths[node] = self.visited_all_paths[node] + fathers_context

        if self.key not in node.context:
            node.context[self.key] = fathers_context

        # include ops from internal function calls
        internal_ops = []
        for ir in node.internal_calls:
            if isinstance(ir.function, Function):
                internal_ops += ir.function.all_slithir_operations()

        # analyze node for retryable tickets
        for ir in node.irs + internal_ops:
            if (
                isinstance(ir, HighLevelCall)
                and isinstance(ir.function, Function)
                and ir.function.name
                in [
                    "createRetryableTicket",
                    "outboundTransferCustomRefund",
                    "unsafeCreateRetryableTicket",
                ]
            ):
                node.context[self.key].append(node)

        if len(node.context[self.key]) > 1:
            self.results.append(node.context[self.key])
            return

        for son in node.sons:
            self._detect_multiple_tickets(function, son, visited)

    def _detect(self) -> list[Output]:
        results = []

        self.results = []
        self.visited_all_paths = {}

        for contract in self.compilation_unit.contracts:
            for function in contract.functions:
                if (
                    function.is_implemented
                    and function.contract_declarer == contract
                    and function.entry_point
                ):
                    function.entry_point.context[self.key] = []
                    self._detect_multiple_tickets(function, function.entry_point, [])

        for multiple_tickets in self.results:
            info = ["Multiple retryable tickets created in the same function:\n"]

            for x in multiple_tickets:
                info += ["\t -", x, "\n"]

            json = self.generate_result(info)
            results.append(json)

        return results



"""
Module detecting ABIEncoderV2 array bug
"""

from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    make_solc_versions,
    DETECTOR_INFO,
)
from slither.core.solidity_types import ArrayType
from slither.core.solidity_types import UserDefinedType
from slither.core.variables.local_variable import LocalVariable
from slither.core.variables.state_variable import StateVariable
from slither.slithir.operations import SolidityCall
from slither.core.declarations.solidity_variables import SolidityFunction
from slither.slithir.operations import EventCall
from slither.slithir.operations import HighLevelCall
from slither.utils.utils import unroll
from slither.core.cfg.node import Node
from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.utils.output import Output


class ABIEncoderV2Array(AbstractDetector):
    """
    Detects Storage ABIEncoderV2 array bug
    """

    ARGUMENT = "abiencoderv2-array"
    HELP = "Storage abiencoderv2 array"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#storage-abiencoderv2-array"
    )
    WIKI_TITLE = "Storage ABIEncoderV2 Array"
    WIKI_DESCRIPTION = """`solc` versions `0.4.7`-`0.5.9` contain a [compiler bug](https://blog.ethereum.org/2019/06/25/solidity-storage-array-bugs) leading to incorrect ABI encoder usage."""

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract A {
    uint[2][3] bad_arr = [[1, 2], [3, 4], [5, 6]];

    /* Array of arrays passed to abi.encode is vulnerable */
    function bad() public {
        bytes memory b = abi.encode(bad_arr);
    }
}
```
`abi.encode(bad_arr)` in a call to `bad()` will incorrectly encode the array as `[[1, 2], [2, 3], [3, 4]]` and lead to unintended behavior.
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Use a compiler >= `0.5.10`."

    VULNERABLE_SOLC_VERSIONS = make_solc_versions(4, 7, 25) + make_solc_versions(5, 0, 9)

    @staticmethod
    def _detect_storage_abiencoderv2_arrays(
        contract: Contract,
    ) -> set[tuple[FunctionContract, Node]]:
        """
        Detects and returns all nodes with storage-allocated abiencoderv2 arrays of arrays/structs in abi.encode, events or external calls
        :param contract: Contract to detect within
        :return: A list of tuples with (function, node) where function node has storage-allocated abiencoderv2 arrays of arrays/structs
        """
        # Create our result set.
        results = set()

        # Loop for each function and modifier.

        for function in contract.functions_and_modifiers_declared:
            # Loop every node, looking for storage-allocated array of arrays/structs
            # in arguments to abi.encode, events or external calls
            for node in function.nodes:
                for ir in node.irs:
                    # Call to abi.encode()
                    if (
                        (
                            isinstance(ir, SolidityCall)
                            and ir.function == SolidityFunction("abi.encode()")
                        )
                        or
                        # Call to emit event
                        # Call to external function
                        isinstance(ir, (EventCall, HighLevelCall))
                    ):
                        for arg in unroll(ir.arguments):
                            # Check if arguments are storage allocated arrays of arrays/structs
                            if (
                                isinstance(arg.type, ArrayType)
                                # Storage allocated
                                and (
                                    isinstance(arg, StateVariable)
                                    or (isinstance(arg, LocalVariable) and arg.is_storage)
                                )
                                # Array of arrays or structs
                                and isinstance(arg.type.type, (ArrayType, UserDefinedType))
                            ):
                                results.add((function, node))
                                break

        # Return the resulting set of tuples
        return results

    def _detect(self) -> list[Output]:
        """
        Detect ABIEncoderV2 array bug
        """
        results = []

        # Check if pragma experimental ABIEncoderV2 is used
        if not any(
            (p.directive[0] == "experimental" and p.directive[1] == "ABIEncoderV2")
            for p in self.compilation_unit.pragma_directives
        ):
            return results

        # Check for storage-allocated abiencoderv2 arrays of arrays/structs
        # in arguments of abi.encode, events or external calls
        for contract in self.contracts:
            storage_abiencoderv2_arrays = self._detect_storage_abiencoderv2_arrays(contract)
            for function, node in storage_abiencoderv2_arrays:
                info: DETECTOR_INFO = [
                    "Function ",
                    function,
                    " trigger an abi encoding bug:\n\t- ",
                    node,
                    "\n",
                ]
                res = self.generate_result(info)
                results.append(res)

        return results



"""
Using For Top Level module
"""

import logging
from typing import TYPE_CHECKING

from slither.core.compilation_unit import SlitherCompilationUnit
from slither.core.declarations import (
    StructureTopLevel,
    EnumTopLevel,
)
from slither.core.declarations.using_for_top_level import UsingForTopLevel
from slither.core.scope.scope import FileScope
from slither.core.solidity_types import TypeAliasTopLevel
from slither.core.solidity_types.user_defined_type import UserDefinedType
from slither.solc_parsing.declarations.caller_context import CallerContextExpression
from slither.solc_parsing.solidity_types.type_parsing import parse_type

if TYPE_CHECKING:
    from slither.solc_parsing.slither_compilation_unit_solc import SlitherCompilationUnitSolc

LOGGER = logging.getLogger("UsingForTopLevelSolc")


class UsingForTopLevelSolc(CallerContextExpression):
    """
    UsingFor class
    """

    def __init__(
        self,
        uftl: UsingForTopLevel,
        top_level_data: dict,
        slither_parser: "SlitherCompilationUnitSolc",
    ) -> None:
        self._type_name = top_level_data["typeName"]
        self._global = top_level_data["global"]

        if "libraryName" in top_level_data:
            self._library_name = top_level_data["libraryName"]
        else:
            self._functions = top_level_data["functionList"]
            self._library_name = None

        self._using_for = uftl
        self._slither_parser = slither_parser

    def analyze(self) -> None:
        type_name = parse_type(self._type_name, self)
        self._using_for.using_for[type_name] = []

        if self._library_name:
            library_name = parse_type(self._library_name, self)
            self._using_for.using_for[type_name].append(library_name)
            self._propagate_global(type_name)
        else:
            for f in self._functions:
                # User defined operator
                if "operator" in f:
                    # Top level function
                    function_name: str = f["definition"]["name"]
                    operator: str = f["operator"]
                    self._analyze_operator(operator, function_name, type_name)
                else:
                    full_name_split = f["function"]["name"].split(".")
                    if len(full_name_split) == 1:
                        # Top level function
                        function_name: str = full_name_split[0]
                        self._analyze_top_level_function(function_name, type_name)
                    elif len(full_name_split) == 2:
                        # It can be a top level function behind an aliased import
                        # or a library function
                        first_part = full_name_split[0]
                        function_name = full_name_split[1]
                        self._check_aliased_import(first_part, function_name, type_name)
                    else:
                        # MyImport.MyLib.a we don't care of the alias
                        library_name_str = full_name_split[1]
                        function_name = full_name_split[2]
                        self._analyze_library_function(library_name_str, function_name, type_name)

    def _check_aliased_import(
        self,
        first_part: str,
        function_name: str,
        type_name: TypeAliasTopLevel | UserDefinedType,
    ) -> None:
        # We check if the first part appear as alias for an import
        # if it is then function_name must be a top level function
        # otherwise it's a library function
        for i in self._using_for.file_scope.imports:
            if i.alias == first_part:
                self._analyze_top_level_function(function_name, type_name)
                return
        self._analyze_library_function(first_part, function_name, type_name)

    def _analyze_top_level_function(
        self, function_name: str, type_name: TypeAliasTopLevel | UserDefinedType
    ) -> None:
        for tl_function in self._using_for.file_scope.functions:
            # The library function is bound to the first parameter's type
            if (
                tl_function.name == function_name
                and tl_function.parameters
                and type_name == tl_function.parameters[0].type
            ):
                self._using_for.using_for[type_name].append(tl_function)
                self._propagate_global(type_name)
                break

    def _analyze_operator(
        self, operator: str, function_name: str, type_name: TypeAliasTopLevel
    ) -> None:
        for tl_function in self._using_for.file_scope.functions:
            # The library function is bound to the first parameter's type
            if (
                tl_function.name == function_name
                and tl_function.parameters
                and type_name == tl_function.parameters[0].type
            ):
                type_name.operators[operator] = tl_function
                break

    def _analyze_library_function(
        self,
        library_name: str,
        function_name: str,
        type_name: TypeAliasTopLevel | UserDefinedType,
    ) -> None:
        found = False
        for c in self.compilation_unit.contracts:
            if found:
                break
            if c.name == library_name:
                for cf in c.functions:
                    # The library function is bound to the first parameter's type
                    if (
                        cf.name == function_name
                        and cf.parameters
                        and type_name == cf.parameters[0].type
                    ):
                        self._using_for.using_for[type_name].append(cf)
                        self._propagate_global(type_name)
                        found = True
                        break
        if not found:
            LOGGER.warning(
                f"Top level using for: Library {library_name} - function {function_name} not found"
            )

    def _propagate_global(self, type_name: TypeAliasTopLevel | UserDefinedType) -> None:
        if self._global:
            for scope in self.compilation_unit.scopes.values():
                if isinstance(type_name, TypeAliasTopLevel):
                    for alias in scope.type_aliases.values():
                        if alias == type_name:
                            scope.using_for_directives.add(self._using_for)
                elif isinstance(type_name, UserDefinedType):
                    self._propagate_global_UserDefinedType(scope, type_name)
                else:
                    LOGGER.error(
                        f"Error when propagating global using for {type_name} {type(type_name)}"
                    )

    def _propagate_global_UserDefinedType(
        self, scope: FileScope, type_name: UserDefinedType
    ) -> None:
        underlying = type_name.type
        if isinstance(underlying, StructureTopLevel):
            for struct in scope.structures.values():
                if struct == underlying:
                    scope.using_for_directives.add(self._using_for)
        elif isinstance(underlying, EnumTopLevel):
            for enum in scope.enums.values():
                if enum == underlying:
                    scope.using_for_directives.add(self._using_for)
        else:
            LOGGER.error(
                f"Error when propagating global {underlying} {type(underlying)} not a StructTopLevel or EnumTopLevel"
            )

    @property
    def is_compact_ast(self) -> bool:
        return self._slither_parser.is_compact_ast

    @property
    def compilation_unit(self) -> SlitherCompilationUnit:
        return self._slither_parser.compilation_unit

    def get_key(self) -> str:
        return self._slither_parser.get_key()

    @property
    def slither_parser(self) -> "SlitherCompilationUnitSolc":
        return self._slither_parser

    @property
    def underlying_using_for(self) -> UsingForTopLevel:
        return self._using_for



"""
This class is used for the SSA version of slithIR
It is similar to the non-SSA version of slithIR
as the TupleVariable are in SSA form in both version
"""

from slither.slithir.variables.tuple import TupleVariable


class TupleVariableSSA(TupleVariable):
    def __init__(self, t: TupleVariable) -> None:
        super().__init__(t.node, t.index)

        self._non_ssa_version = t

    @property
    def non_ssa_version(self):
        return self._non_ssa_version



"""
Event module
"""

from typing import TYPE_CHECKING, Union

from slither.core.cfg.node import NodeType
from slither.core.cfg.node import link_nodes
from slither.core.cfg.scope import Scope
from slither.core.declarations.modifier import Modifier
from slither.solc_parsing.cfg.node import NodeSolc
from slither.solc_parsing.declarations.function import FunctionSolc

if TYPE_CHECKING:
    from slither.solc_parsing.declarations.contract import ContractSolc
    from slither.solc_parsing.slither_compilation_unit_solc import SlitherCompilationUnitSolc
    from slither.core.declarations import Function


class ModifierSolc(FunctionSolc):
    def __init__(
        self,
        modifier: Modifier,
        function_data: dict,
        contract_parser: "ContractSolc",
        slither_parser: "SlitherCompilationUnitSolc",
    ) -> None:
        super().__init__(modifier, function_data, contract_parser, slither_parser)
        # _modifier is equal to _function, but keep it here to prevent
        # confusion for mypy in underlying_function
        self._modifier = modifier

    @property
    def underlying_function(self) -> Modifier:
        return self._modifier

    def analyze_params(self) -> None:
        # Can be re-analyzed due to inheritance
        if self._params_was_analyzed:
            return

        self._params_was_analyzed = True

        self._analyze_attributes()

        if self.is_compact_ast:
            params = self._functionNotParsed["parameters"]
        else:
            children = self._functionNotParsed["children"]
            # It uses to be
            # params = children[0]
            # But from Solidity 0.6.3 to 0.6.10 (included)
            # Comment above a function might be added in the children
            params = next(child for child in children if child[self.get_key()] == "ParameterList")

        if params:
            self._parse_params(params)

    def analyze_content(self) -> None:
        if self._content_was_analyzed:
            return

        self._content_was_analyzed = True

        if self.is_compact_ast:
            body = self._functionNotParsed.get("body", None)

            if body and body[self.get_key()] == "Block":
                self._function.is_implemented = True
                self._parse_cfg(body)

        else:
            children = self._functionNotParsed["children"]

            self._function.is_implemented = False
            if len(children) > 1:
                # It uses to be
                # params = children[1]
                # But from Solidity 0.6.3 to 0.6.10 (included)
                # Comment above a function might be added in the children
                block = next(child for child in children if child[self.get_key()] == "Block")
                self._function.is_implemented = True
                self._parse_cfg(block)

        for local_var_parser in self._local_variables_parser:
            local_var_parser.analyze(self)

        for node in self._node_to_nodesolc.values():
            node.analyze_expressions(self)

        for yul_parser in self._node_to_yulobject.values():
            yul_parser.analyze_expressions()

        self._rewrite_ternary_as_if_else()
        self._remove_alone_endif()

        if self._function.entry_point:
            self._update_reachability(self._function.entry_point)
        # self._analyze_read_write()
        # self._analyze_calls()

    def _parse_statement(
        self, statement: dict, node: NodeSolc, scope: Union[Scope, "Function"]
    ) -> NodeSolc:
        name = statement[self.get_key()]
        if name == "PlaceholderStatement":
            placeholder_node = self._new_node(NodeType.PLACEHOLDER, statement["src"], scope)
            link_nodes(node.underlying_node, placeholder_node.underlying_node)
            return placeholder_node
        return super()._parse_statement(statement, node, scope)



import re
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.formatters.naming_convention.naming_convention import custom_format
from slither.utils.output import Output


class NamingConvention(AbstractDetector):
    """
    Check if naming conventions are followed
    https://solidity.readthedocs.io/en/v0.4.25/style-guide.html?highlight=naming_convention%20convention#naming_convention-conventions

    Exceptions:
    - Allow constant variables name/symbol/decimals to be lowercase (ERC20)
    - Allow '_' at the beginning of the mixed_case match for private variables and unused parameters
    - Ignore echidna properties (functions with names starting 'echidna_' or 'crytic_'
    """

    ARGUMENT = "naming-convention"
    HELP = "Conformity to Solidity naming conventions"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH
    LANGUAGE = "solidity"
    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#conformance-to-solidity-naming-conventions"

    WIKI_TITLE = "Conformance to Solidity naming conventions"

    # region wiki_description
    WIKI_DESCRIPTION = """
Solidity defines a [naming convention](https://solidity.readthedocs.io/en/v0.4.25/style-guide.html#naming-conventions) that should be followed.
#### Rule exceptions
- Allow constant variable name/symbol/decimals to be lowercase (`ERC20`).
- Allow `_` at the beginning of the `mixed_case` match for private variables and unused parameters."""
    # endregion wiki_description

    WIKI_RECOMMENDATION = "Follow the Solidity [naming convention](https://solidity.readthedocs.io/en/v0.4.25/style-guide.html#naming-conventions)."

    STANDARD_JSON = False

    @staticmethod
    def is_cap_words(name: str) -> bool:
        return re.search("^[A-Z]([A-Za-z0-9]+)?_?$", name) is not None

    @staticmethod
    def is_immutable_naming(name: str) -> bool:
        return re.search("^i_[a-z]([A-Za-z0-9]+)?_?$", name) is not None

    @staticmethod
    def is_state_naming(name: str) -> bool:
        return re.search("^s_[a-z]([A-Za-z0-9]+)?_?$", name) is not None

    @staticmethod
    def is_mixed_case(name: str) -> bool:
        return re.search("^[a-z]([A-Za-z0-9]+)?_?$", name) is not None

    @staticmethod
    def is_mixed_case_with_underscore(name: str) -> bool:
        # Allow _ at the beginning to represent private variable
        # or unused parameters
        return re.search("^[_]?[a-z]([A-Za-z0-9]+)?_?$", name) is not None

    @staticmethod
    def is_upper_case_with_underscores(name: str) -> bool:
        return re.search("^[A-Z0-9_]+_?$", name) is not None

    @staticmethod
    def should_avoid_name(name: str) -> bool:
        return re.search("^[lOI]$", name) is not None

    def _detect(self) -> list[Output]:
        results = []
        info: DETECTOR_INFO
        for contract in self.contracts:
            if not self.is_cap_words(contract.name):
                info = ["Contract ", contract, " is not in CapWords\n"]

                res = self.generate_result(info)
                res.add(contract, {"target": "contract", "convention": "CapWords"})
                results.append(res)

            for struct in contract.structures_declared:
                if not self.is_cap_words(struct.name):
                    info = ["Struct ", struct, " is not in CapWords\n"]

                    res = self.generate_result(info)
                    res.add(struct, {"target": "structure", "convention": "CapWords"})
                    results.append(res)

            for event in contract.events_declared:
                if not self.is_cap_words(event.name):
                    info = ["Event ", event, " is not in CapWords\n"]

                    res = self.generate_result(info)
                    res.add(event, {"target": "event", "convention": "CapWords"})
                    results.append(res)

            for func in contract.functions_declared:
                if func.is_constructor:
                    continue
                if not self.is_mixed_case(func.name):
                    if func.visibility in [
                        "internal",
                        "private",
                    ] and self.is_mixed_case_with_underscore(func.name):
                        continue
                    if func.name.startswith(("echidna_", "crytic_")):
                        continue
                    info = ["Function ", func, " is not in mixedCase\n"]

                    res = self.generate_result(info)
                    res.add(func, {"target": "function", "convention": "mixedCase"})
                    results.append(res)

                for argument in func.parameters:
                    # Ignore parameter names that are not specified i.e. empty strings
                    if argument.name == "":
                        continue
                    if argument in func.variables_read_or_written:
                        correct_naming = self.is_mixed_case(argument.name)
                    else:
                        correct_naming = self.is_mixed_case_with_underscore(argument.name)
                    if not correct_naming:
                        info = ["Parameter ", argument, " is not in mixedCase\n"]

                        res = self.generate_result(info)
                        res.add(argument, {"target": "parameter", "convention": "mixedCase"})
                        results.append(res)

            for var in contract.state_variables_declared:
                if self.should_avoid_name(var.name):
                    info = [
                        "Variable ",
                        var,
                        " is single letter l, O, or I, which should not be used\n",
                    ]

                    res = self.generate_result(info)
                    res.add(
                        var,
                        {
                            "target": "variable",
                            "convention": "l_O_I_should_not_be_used",
                        },
                    )
                    results.append(res)

                if var.is_constant is True:
                    # For ERC20 compatibility
                    if var.name in ["symbol", "name", "decimals"]:
                        continue
                    if var.visibility == "public":
                        continue
                    if not self.is_upper_case_with_underscores(var.name):
                        info = [
                            "Constant ",
                            var,
                            " is not in UPPER_CASE_WITH_UNDERSCORES\n",
                        ]

                        res = self.generate_result(info)
                        res.add(
                            var,
                            {
                                "target": "variable_constant",
                                "convention": "UPPER_CASE_WITH_UNDERSCORES",
                            },
                        )
                        results.append(res)

                else:
                    if var.visibility in ["private", "internal"]:
                        correct_naming = self.is_mixed_case_with_underscore(
                            var.name
                        ) or self.is_state_naming(var.name)

                        if not correct_naming and var.is_immutable:
                            correct_naming = self.is_immutable_naming(var.name)

                    else:
                        correct_naming = self.is_mixed_case(var.name)

                    if not correct_naming:
                        info = ["Variable ", var, " is not in mixedCase\n"]

                        res = self.generate_result(info)
                        res.add(var, {"target": "variable", "convention": "mixedCase"})
                        results.append(res)

            for enum in contract.enums_declared:
                if not self.is_cap_words(enum.name):
                    info = ["Enum ", enum, " is not in CapWords\n"]

                    res = self.generate_result(info)
                    res.add(enum, {"target": "enum", "convention": "CapWords"})
                    results.append(res)

            for modifier in contract.modifiers_declared:
                if not self.is_mixed_case(modifier.name):
                    info = ["Modifier ", modifier, " is not in mixedCase\n"]

                    res = self.generate_result(info)
                    res.add(modifier, {"target": "modifier", "convention": "mixedCase"})
                    results.append(res)

        return results

    @staticmethod
    def _format(slither, result):
        custom_format(slither, result)



"""
Detects various properties of inheritance in provided contracts.
"""

from collections import defaultdict
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from slither.core.declarations import Contract, Function
    from slither.core.variables.state_variable import StateVariable


def detect_c3_function_shadowing(
    contract: "Contract",
) -> dict["Function", set["Function"]]:
    """
    Detects and obtains functions which are indirectly shadowed via multiple inheritance by C3 linearization
    properties, despite not directly inheriting from each other.

    :param contract: The contract to check for potential C3 linearization shadowing within.
    :return: A dict (function winner -> [shadowed functions])
    """

    targets: dict[str, Function] = {
        f.full_name: f
        for f in contract.functions_inherited
        if f.shadows and not f.is_constructor and not f.is_constructor_variables
    }

    collisions: dict[Function, set[Function]] = defaultdict(set)

    for contract_inherited in contract.immediate_inheritance:
        for candidate in contract_inherited.functions:
            if candidate.full_name not in targets or candidate.is_shadowed:
                continue
            if targets[candidate.full_name].canonical_name != candidate.canonical_name:
                collisions[targets[candidate.full_name]].add(candidate)
    return collisions


def detect_state_variable_shadowing(
    contracts: list["Contract"],
) -> set[tuple["Contract", "StateVariable", "Contract", "StateVariable"]]:
    """
    Detects all overshadowing and overshadowed state variables in the provided contracts.
    :param contracts: The contracts to detect shadowing within.
    :return: Returns a set of tuples (overshadowing_contract, overshadowing_state_var, overshadowed_contract,
    overshadowed_state_var).
    The contract-variable pair's variable does not need to be defined in its paired contract, it may have been
    inherited. The contracts are simply included to denote the immediate inheritance path from which the shadowed
    variable originates.
    """
    results: set[tuple[Contract, StateVariable, Contract, StateVariable]] = set()
    for contract in contracts:
        variables_declared: dict[str, StateVariable] = {
            variable.name: variable for variable in contract.state_variables_declared
        }
        for immediate_base_contract in contract.immediate_inheritance:
            for variable in immediate_base_contract.variables:
                if variable.name in variables_declared:
                    results.add(
                        (
                            contract,
                            variables_declared[variable.name],
                            immediate_base_contract,
                            variable,
                        )
                    )
    return results



from slither.core.cfg.node import Node
from slither.core.declarations import Function, SolidityVariable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class VarReadUsingThis(AbstractDetector):
    ARGUMENT = "var-read-using-this"
    HELP = "Contract reads its own variable using `this`"
    IMPACT = DetectorClassification.OPTIMIZATION
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#public-variable-read-in-external-context"

    WIKI_TITLE = "Public variable read in external context"
    WIKI_DESCRIPTION = "The contract reads its own variable using `this`, adding overhead of an unnecessary STATICCALL."
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract C {
    mapping(uint => address) public myMap;
    function test(uint x) external returns(address) {
        return this.myMap(x);
    }
}
```
"""

    WIKI_RECOMMENDATION = "Read the variable directly from storage instead of calling the contract."

    def _detect(self) -> list[Output]:
        results = []
        for c in self.contracts:
            for func in c.functions:
                for node in self._detect_var_read_using_this(func):
                    info: DETECTOR_INFO = [
                        "The function ",
                        func,
                        " reads ",
                        node,
                        " with `this` which adds an extra STATICCALL.\n",
                    ]
                    json = self.generate_result(info)
                    results.append(json)

        return results

    @staticmethod
    def _detect_var_read_using_this(func: Function) -> list[Node]:
        results: list[Node] = []
        for _, ir in func.high_level_calls:
            if (
                ir.destination == SolidityVariable("this")
                and ir.is_static_call()
                and ir.function.visibility == "public"
            ):
                results.append(ir.node)
        return sorted(results, key=lambda x: x.node_id)



from slither.utils.output import Output
from slither.core.compilation_unit import SlitherCompilationUnit
from slither.formatters.variables.unchanged_state_variables import custom_format
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from .unchanged_state_variables import UnchangedStateVariables


class CouldBeImmutable(AbstractDetector):
    """
    State variables that could be declared immutable.
    # Immutable attribute available in Solidity 0.6.5 and above
    # https://blog.soliditylang.org/2020/04/06/solidity-0.6.5-release-announcement/
    """

    # VULNERABLE_SOLC_VERSIONS =
    ARGUMENT = "immutable-states"
    HELP = "State variables that could be declared immutable"
    IMPACT = DetectorClassification.OPTIMIZATION
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#state-variables-that-could-be-declared-immutable"

    WIKI_TITLE = "State variables that could be declared immutable"
    WIKI_DESCRIPTION = "State variables that are not updated following deployment should be declared immutable to save gas."
    WIKI_RECOMMENDATION = "Add the `immutable` attribute to state variables that never change or are set only in the constructor."

    def _detect(self) -> list[Output]:
        """Detect state variables that could be immutable"""
        results = {}
        unchanged_state_variables = UnchangedStateVariables(self.compilation_unit)
        unchanged_state_variables.detect()

        for variable in unchanged_state_variables.immutable_candidates:
            results[variable.canonical_name] = self.generate_result(
                [variable, " should be immutable \n"]
            )

        # Order by canonical name for deterministic results
        return [results[k] for k in sorted(results)]

    @staticmethod
    def _format(compilation_unit: SlitherCompilationUnit, result: dict) -> None:
        custom_format(compilation_unit, result, "immutable")



from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class ChainlinkFeedRegistry(AbstractDetector):
    ARGUMENT = "chainlink-feed-registry"
    HELP = "Detect when chainlink feed registry is used"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#chainlink-feed-registry-usage"

    WIKI_TITLE = "Chainlink Feed Registry usage"
    WIKI_DESCRIPTION = "Detect when Chainlink Feed Registry is used. At the moment is only available on Ethereum Mainnet."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
import "chainlink/contracts/src/v0.8/interfaces/FeedRegistryInteface.sol"

contract A {
    FeedRegistryInterface public immutable registry;

    constructor(address _registry) {
        registry = _registry;
    }

    function getPrice(address base, address quote) public return(uint256) {
        (, int256 price,,,) = registry.latestRoundData(base, quote);
        // Do price validation
        return uint256(price);
    }
}
```
If the contract is deployed on a different chain than Ethereum Mainnet the `getPrice` function will revert.
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Do not use Chainlink Feed Registry outside of Ethereum Mainnet."

    def _detect(self) -> list[Output]:
        # https://github.com/smartcontractkit/chainlink/blob/8ca41fc8f722accfccccb4b1778db2df8fef5437/contracts/src/v0.8/interfaces/FeedRegistryInterface.sol
        registry_functions = [
            "decimals",
            "description",
            "versiom",
            "latestRoundData",
            "getRoundData",
            "latestAnswer",
            "latestTimestamp",
            "latestRound",
            "getAnswer",
            "getTimestamp",
            "getFeed",
            "getPhaseFeed",
            "isFeedEnabled",
            "getPhase",
            "getRoundFeed",
            "getPhaseRange",
            "getPreviousRoundId",
            "getNextRoundId",
            "proposeFeed",
            "confirmFeed",
            "getProposedFeed",
            "proposedGetRoundData",
            "proposedLatestRoundData",
            "getCurrentPhaseId",
        ]
        results = []

        for contract in self.compilation_unit.contracts_derived:
            nodes = []
            for target, ir in contract.all_high_level_calls:
                if (
                    target.name == "FeedRegistryInterface"
                    and ir.function_name in registry_functions
                ):
                    nodes.append(ir.node)
            # Sort so output is deterministic
            nodes.sort(key=lambda x: (x.node_id, x.function.full_name))

            if len(nodes) > 0:
                info: DETECTOR_INFO = [
                    "The Chainlink Feed Registry is used in the ",
                    contract.name,
                    " contract. It's only available on Ethereum Mainnet, consider to not use it if the contract needs to be deployed on other chains.\n",
                ]

                for node in nodes:
                    info.extend(["\t - ", node, "\n"])

                res = self.generate_result(info)
                results.append(res)

        return results



"""
CK Metrics are a suite of six software metrics proposed by Chidamber and Kemerer in 1994.
These metrics are used to measure the complexity of a class.
https://en.wikipedia.org/wiki/Programming_complexity

- Response For a Class (RFC) is a metric that measures the number of unique method calls within a class.
- Number of Children (NOC) is a metric that measures the number of children a class has.
- Depth of Inheritance Tree (DIT) is a metric that measures the number of parent classes a class has.
- Coupling Between Object Classes (CBO) is a metric that measures the number of classes a class is coupled to.

Not implemented:
- Lack of Cohesion of Methods (LCOM) is a metric that measures the lack of cohesion in methods.
- Weighted Methods per Class (WMC) is a metric that measures the complexity of a class.

During the calculation of the metrics above, there are a number of other intermediate metrics that are calculated.
These are also included in the output:
 - State variables: total number of state variables
 - Constants: total number of constants
 - Immutables: total number of immutables
 - Public: total number of public functions
 - External: total number of external functions
 - Internal: total number of internal functions
 - Private: total number of private functions
 - Mutating: total number of state mutating functions
 - View: total number of view functions
 - Pure: total number of pure functions
 - External mutating: total number of external mutating functions
 - No auth or onlyOwner: total number of functions without auth or onlyOwner modifiers
 - No modifiers: total number of functions without modifiers
 - Ext calls: total number of external calls

"""

from collections import OrderedDict
from dataclasses import dataclass, field
from slither.utils.colors import bold
from slither.core.declarations import Contract
from slither.utils.myprettytable import make_pretty_table, MyPrettyTable
from slither.utils.martin import MartinMetrics
from slither.slithir.operations.high_level_call import HighLevelCall


# Utility functions


def compute_dit(contract: Contract, depth: int = 0) -> int:
    """
    Recursively compute the depth of inheritance tree (DIT) of a contract
    Args:
        contract(core.declarations.contract.Contract): contract to compute DIT for
        depth(int): current depth of the contract
    Returns:
        int: depth of the contract
    """
    if not contract.inheritance:
        return depth
    max_dit = depth
    for inherited_contract in contract.inheritance:
        dit = compute_dit(inherited_contract, depth + 1)
        max_dit = max(max_dit, dit)
    return max_dit


def has_auth(func) -> bool:
    """
    Check if a function has no auth or only_owner modifiers
    Args:
        func(core.declarations.function.Function): function to check
    Returns:
        bool True if it does have auth or only_owner modifiers
    """
    for modifier in func.modifiers:
        if "auth" in modifier.name or "only_owner" in modifier.name:
            return True
    return False


# Utility classes for calculating CK metrics


@dataclass(slots=True)
class CKContractMetrics:
    """Class to hold the CK metrics for a single contract."""

    contract: Contract

    # Used to calculate CBO - should be passed in as a constructor arg
    martin_metrics: dict

    # Used to calculate NOC
    dependents: dict

    state_variables: int = 0
    constants: int = 0
    immutables: int = 0
    public: int = 0
    external: int = 0
    internal: int = 0
    private: int = 0
    mutating: int = 0
    view: int = 0
    pure: int = 0
    external_mutating: int = 0
    no_auth_or_only_owner: int = 0
    no_modifiers: int = 0
    ext_calls: int = 0
    rfc: int = 0
    noc: int = 0
    dit: int = 0
    cbo: int = 0

    def __post_init__(self) -> None:
        if not hasattr(self.contract, "functions"):
            return
        self.count_variables()
        self.noc = len(self.dependents[self.contract.name])
        self.dit = compute_dit(self.contract)
        self.cbo = (
            self.martin_metrics[self.contract.name].ca + self.martin_metrics[self.contract.name].ce
        )
        self.calculate_metrics()

    def calculate_metrics(self) -> None:
        """Calculate the metrics for a contract"""
        rfc = self.public  # initialize with public getter count
        for func in self.contract.functions:
            if func.name == "constructor":
                continue
            pure = func.pure
            view = not pure and func.view
            mutating = not pure and not view
            external = func.visibility == "external"
            public = func.visibility == "public"
            internal = func.visibility == "internal"
            private = func.visibility == "private"
            external_public_mutating = external or (public and mutating)
            external_no_auth = external_public_mutating and not has_auth(func)
            external_no_modifiers = external_public_mutating and len(func.modifiers) == 0
            if external or public:
                rfc += 1

            high_level_calls = [
                ir for node in func.nodes for ir in node.irs_ssa if isinstance(ir, HighLevelCall)
            ]

            # convert irs to string with target function and contract name
            external_calls = []
            for high_level_call in high_level_calls:
                if isinstance(high_level_call.destination, Contract):
                    destination_contract = high_level_call.destination.name
                elif isinstance(high_level_call.destination, str):
                    destination_contract = high_level_call.destination
                elif not hasattr(high_level_call.destination, "type"):
                    continue
                elif isinstance(high_level_call.destination.type, Contract):
                    destination_contract = high_level_call.destination.type.name
                elif isinstance(high_level_call.destination.type, str):
                    destination_contract = high_level_call.destination.type
                elif not hasattr(high_level_call.destination.type, "type"):
                    continue
                elif isinstance(high_level_call.destination.type.type, Contract):
                    destination_contract = high_level_call.destination.type.type.name
                elif isinstance(high_level_call.destination.type.type, str):
                    destination_contract = high_level_call.destination.type.type
                else:
                    continue
                external_calls.append(f"{high_level_call.function_name}{destination_contract}")
            rfc += len(set(external_calls))

            self.public += public
            self.external += external
            self.internal += internal
            self.private += private

            self.mutating += mutating
            self.view += view
            self.pure += pure

            self.external_mutating += external_public_mutating
            self.no_auth_or_only_owner += external_no_auth
            self.no_modifiers += external_no_modifiers

            self.ext_calls += len(external_calls)
            self.rfc = rfc

    def count_variables(self) -> None:
        """Count the number of variables in a contract"""
        state_variable_count = 0
        constant_count = 0
        immutable_count = 0
        public_getter_count = 0
        for variable in self.contract.variables:
            if variable.is_constant:
                constant_count += 1
            elif variable.is_immutable:
                immutable_count += 1
            else:
                state_variable_count += 1
            if variable.visibility == "Public":
                public_getter_count += 1
        self.state_variables = state_variable_count
        self.constants = constant_count
        self.immutables = immutable_count

        # initialize RFC with public getter count
        # self.public is used count public functions not public variables
        self.rfc = public_getter_count

    def to_dict(self) -> dict[str, float]:
        """Return the metrics as a dictionary."""
        return OrderedDict(
            {
                "State variables": self.state_variables,
                "Constants": self.constants,
                "Immutables": self.immutables,
                "Public": self.public,
                "External": self.external,
                "Internal": self.internal,
                "Private": self.private,
                "Mutating": self.mutating,
                "View": self.view,
                "Pure": self.pure,
                "External mutating": self.external_mutating,
                "No auth or onlyOwner": self.no_auth_or_only_owner,
                "No modifiers": self.no_modifiers,
                "Ext calls": self.ext_calls,
                "RFC": self.rfc,
                "NOC": self.noc,
                "DIT": self.dit,
                "CBO": self.cbo,
            }
        )


@dataclass(slots=True)
class SectionInfo:
    """Class to hold the information for a section of the report."""

    title: str
    pretty_table: MyPrettyTable
    txt: str


@dataclass(slots=True)
class CKMetrics:
    """Class to hold the CK metrics for all contracts. Contains methods useful for reporting.

    There are 5 sections in the report:
    1. Variable count by type (state, constant, immutable)
    2. Function count by visibility (public, external, internal, private)
    3. Function count by mutability (mutating, view, pure)
    4. External mutating function count by modifier (external mutating, no auth or onlyOwner, no modifiers)
    5. CK metrics (RFC, NOC, DIT, CBO)
    """

    contracts: list[Contract] = field(default_factory=list)
    contract_metrics: OrderedDict = field(default_factory=OrderedDict)
    title: str = "CK complexity metrics"
    full_text: str = ""
    auxiliary1: SectionInfo = field(default=SectionInfo)
    auxiliary2: SectionInfo = field(default=SectionInfo)
    auxiliary3: SectionInfo = field(default=SectionInfo)
    auxiliary4: SectionInfo = field(default=SectionInfo)
    core: SectionInfo = field(default=SectionInfo)
    AUXILIARY1_KEYS = (
        "State variables",
        "Constants",
        "Immutables",
    )
    AUXILIARY2_KEYS = (
        "Public",
        "External",
        "Internal",
        "Private",
    )
    AUXILIARY3_KEYS = (
        "Mutating",
        "View",
        "Pure",
    )
    AUXILIARY4_KEYS = (
        "External mutating",
        "No auth or onlyOwner",
        "No modifiers",
    )
    CORE_KEYS = (
        "Ext calls",
        "RFC",
        "NOC",
        "DIT",
        "CBO",
    )
    SECTIONS: tuple[tuple[str, str, tuple[str]]] = (
        ("Variables", "auxiliary1", AUXILIARY1_KEYS),
        ("Function visibility", "auxiliary2", AUXILIARY2_KEYS),
        ("State mutability", "auxiliary3", AUXILIARY3_KEYS),
        ("External mutating functions", "auxiliary4", AUXILIARY4_KEYS),
        ("Core", "core", CORE_KEYS),
    )

    def __post_init__(self) -> None:
        martin_metrics = MartinMetrics(self.contracts).contract_metrics
        dependents = {
            inherited.name: {
                contract.name
                for contract in self.contracts
                if inherited.name in contract.inheritance
            }
            for inherited in self.contracts
        }
        for contract in self.contracts:
            self.contract_metrics[contract.name] = CKContractMetrics(
                contract=contract, martin_metrics=martin_metrics, dependents=dependents
            )

        # Create the table and text for each section.
        data = {
            contract.name: self.contract_metrics[contract.name].to_dict()
            for contract in self.contracts
        }

        subtitle = ""
        # Update each section
        for title, attr, keys in self.SECTIONS:
            if attr == "core":
                # Special handling for core section
                totals_enabled = False
                subtitle += bold("RFC: Response For a Class\n")
                subtitle += bold("NOC: Number of Children\n")
                subtitle += bold("DIT: Depth of Inheritance Tree\n")
                subtitle += bold("CBO: Coupling Between Object Classes\n")
            else:
                totals_enabled = True
                subtitle = ""

            pretty_table = make_pretty_table(["Contract", *keys], data, totals=totals_enabled)
            section_title = f"{self.title} ({title})"
            txt = f"\n\n{section_title}:\n{subtitle}{pretty_table}\n"
            self.full_text += txt
            setattr(
                self,
                attr,
                SectionInfo(title=section_title, pretty_table=pretty_table, txt=txt),
            )



from pathlib import Path
import shutil

import pytest
from slither import Slither

# NB: read tests/source_unit/README.md for setup before using this test


TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"

foundry_available = shutil.which("forge") is not None
project_ready = Path(TEST_DATA_DIR, "test_source_unit/lib/forge-std").exists()


@pytest.mark.skipif(
    not foundry_available or not project_ready, reason="requires Foundry and project setup"
)
def test_contract_info() -> None:
    slither = Slither(Path(TEST_DATA_DIR, "test_source_unit").as_posix())

    assert len(slither.compilation_units) == 1
    compilation_unit = slither.compilation_units[0]

    for source_unit in compilation_unit.crytic_compile_compilation_unit.source_units.values():
        source_unit.remove_metadata()

    counter_sol = compilation_unit.crytic_compile.filename_lookup(
        Path(TEST_DATA_DIR, "test_source_unit/src/Counter.sol").as_posix()
    )
    assert (
        compilation_unit.scopes[counter_sol].bytecode_init(
            compilation_unit.crytic_compile_compilation_unit, "Counter"
        )
        == "608060405234801561001057600080fd5b5060f78061001f6000396000f3fe6080604052348015600f57600080fd5b5060043610603c5760003560e01c80633fb5c1cb1460415780638381f58a146053578063d09de08a14606d575b600080fd5b6051604c3660046083565b600055565b005b605b60005481565b60405190815260200160405180910390f35b6051600080549080607c83609b565b9190505550565b600060208284031215609457600080fd5b5035919050565b60006001820160ba57634e487b7160e01b600052601160045260246000fd5b506001019056fe"
    )

    counter2_sol = compilation_unit.crytic_compile.filename_lookup(
        "tests/source_unit/src/Counter2.sol"
    )
    assert (
        compilation_unit.scopes[counter2_sol].bytecode_init(
            compilation_unit.crytic_compile_compilation_unit, "Counter"
        )
        == "6080604052348015600f57600080fd5b50603f80601d6000396000f3fe6080604052600080fdfe"
    )



import logging

from slither.core.declarations.function import Function
from slither.core.solidity_types import Type
from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.utils.utils import is_valid_lvalue, is_valid_rvalue, RVALUE, LVALUE
from slither.slithir.variables import TupleVariable, ReferenceVariable

logger = logging.getLogger("AssignmentOperationIR")


class Assignment(OperationWithLValue):
    def __init__(
        self,
        left_variable: LVALUE,
        right_variable: RVALUE | Function | TupleVariable,
        variable_return_type: Type,
    ) -> None:
        assert is_valid_lvalue(left_variable)
        assert is_valid_rvalue(right_variable) or isinstance(
            right_variable, (Function, TupleVariable)
        )
        super().__init__()
        self._variables = [left_variable, right_variable]
        self._lvalue: LVALUE = left_variable
        self._rvalue: RVALUE | Function | TupleVariable = right_variable
        self._variable_return_type = variable_return_type

    @property
    def variables(self) -> list[LVALUE | RVALUE | Function | TupleVariable]:
        return list(self._variables)

    @property
    def read(self) -> list[RVALUE | Function | TupleVariable]:
        return [self.rvalue]

    @property
    def variable_return_type(self) -> Type:
        return self._variable_return_type

    @property
    def rvalue(self) -> RVALUE | Function | TupleVariable:
        return self._rvalue

    def __str__(self) -> str:
        lvalue = self.lvalue

        # When rvalues are functions, we want to properly display their return type
        # Fix: https://github.com/crytic/slither/issues/2266
        if isinstance(self.rvalue.type, list):
            rvalue_type = ",".join(f"{rvalue_type}" for rvalue_type in self.rvalue.type)
        else:
            rvalue_type = f"{self.rvalue.type}"

        assert lvalue
        if lvalue and isinstance(lvalue, ReferenceVariable):
            points = lvalue.points_to
            while isinstance(points, ReferenceVariable):
                points = points.points_to
            return f"{lvalue}({lvalue.type}) (->{points}) := {self.rvalue}({rvalue_type})"

        return f"{lvalue}({lvalue.type}) := {self.rvalue}({rvalue_type})"



import sys
from slither import Slither

if len(sys.argv) != 2:
    print("python slithIR.py contract.sol")
    sys.exit(-1)

# Init slither
slither = Slither(sys.argv[1])

# Iterate over all the contracts
for contract in slither.contracts:
    # Iterate over all the functions
    for function in contract.functions:
        # Dont explore inherited functions
        if function.contract_declarer == contract:
            print(f"Function: {function.name}")

            # Iterate over the nodes of the function
            for node in function.nodes:
                # Print the Solidity expression of the nodes
                # And the SlithIR operations
                if node.expression:
                    print(f"\tSolidity expression: {node.expression}")
                    print("\tSlithIR:")
                    for ir in node.irs:
                        print(f"\t\t\t{ir}")



from pathlib import Path
from slither import Slither
from slither.core.declarations.contract import Contract
from slither.slithir.operations import LibraryCall, InternalCall

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"


def test_top_level_using_for(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.24")
    slither = Slither(Path(TEST_DATA_DIR, "top_level_using_for.sol").as_posix(), solc=solc_path)

    function = slither.compilation_units[0].functions_top_level[1]
    assert function.name == "b"

    # LIBRARY_CALL, dest:Lib, function:Lib.a(uint256), arguments:['4']
    first_ir = function.slithir_operations[0]
    assert (
        isinstance(first_ir, LibraryCall)
        and isinstance(first_ir.destination, Contract)
        and first_ir.destination.name == "Lib"
        and first_ir.function_name == "a"
        and len(first_ir.arguments) == 1
    )

    # INTERNAL_CALL, c(uint256)(y)
    second_ir = function.slithir_operations[1]
    assert (
        isinstance(second_ir, InternalCall)
        and second_ir.function_name == "c"
        and len(second_ir.arguments) == 1
        and second_ir.arguments[0].name == "y"
    )

    # LIBRARY_CALL, dest:Lib, function:Lib.a(uint256), arguments:['y']
    third_ir = function.slithir_operations[2]
    assert (
        isinstance(third_ir, LibraryCall)
        and isinstance(third_ir.destination, Contract)
        and third_ir.destination.name == "Lib"
        and third_ir.function_name == "a"
        and len(third_ir.arguments) == 1
        and third_ir.arguments[0].name == "y"
    )



from slither.core.declarations import Contract, Structure, Enum
from slither.core.declarations.using_for_top_level import UsingForTopLevel
from slither.core.solidity_types import (
    UserDefinedType,
    Type,
    ElementaryType,
    TypeAlias,
    MappingType,
    ArrayType,
)
from slither.core.solidity_types.elementary_type import Uint, Int, Byte
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


def _is_correctly_used(type_: Type, library: Contract) -> bool:
    """
    Checks if a `using library for type_` statement is used correctly (that is, does library contain any function
    with type_ as the first argument).
    """
    for f in library.functions:
        if len(f.parameters) == 0:
            continue
        if f.parameters[0].type and not _implicitly_convertible_to(type_, f.parameters[0].type):
            continue
        return True
    return False


def _implicitly_convertible_to(type1: Type, type2: Type) -> bool:
    """
    Returns True if type1 may be implicitly converted to type2.
    """
    if isinstance(type1, TypeAlias) or isinstance(type2, TypeAlias):
        if isinstance(type1, TypeAlias) and isinstance(type2, TypeAlias):
            return type1.type == type2.type
        return False

    if isinstance(type1, UserDefinedType) and isinstance(type2, UserDefinedType):
        if isinstance(type1.type, Contract) and isinstance(type2.type, Contract):
            return _implicitly_convertible_to_for_contracts(type1.type, type2.type)

        if isinstance(type1.type, Structure) and isinstance(type2.type, Structure):
            return type1.type.canonical_name == type2.type.canonical_name

        if isinstance(type1.type, Enum) and isinstance(type2.type, Enum):
            return type1.type.canonical_name == type2.type.canonical_name

    if isinstance(type1, ElementaryType) and isinstance(type2, ElementaryType):
        return _implicitly_convertible_to_for_elementary_types(type1, type2)

    if isinstance(type1, MappingType) and isinstance(type2, MappingType):
        return _implicitly_convertible_to_for_mappings(type1, type2)

    if isinstance(type1, ArrayType) and isinstance(type2, ArrayType):
        return _implicitly_convertible_to_for_arrays(type1, type2)

    return False


def _implicitly_convertible_to_for_arrays(type1: ArrayType, type2: ArrayType) -> bool:
    """
    Returns True if type1 may be implicitly converted to type2.
    """
    return _implicitly_convertible_to(type1.type, type2.type)


def _implicitly_convertible_to_for_mappings(type1: MappingType, type2: MappingType) -> bool:
    """
    Returns True if type1 may be implicitly converted to type2.
    """
    return type1.type_from == type2.type_from and type1.type_to == type2.type_to


def _implicitly_convertible_to_for_elementary_types(
    type1: ElementaryType, type2: ElementaryType
) -> bool:
    """
    Returns True if type1 may be implicitly converted to type2.
    """
    if type1.type == "bool" and type2.type == "bool":
        return True
    if type1.type == "string" and type2.type == "string":
        return True
    if type1.type == "bytes" and type2.type == "bytes":
        return True
    if type1.type == "address" and type2.type == "address":
        return _implicitly_convertible_to_for_addresses(type1, type2)
    if type1.type in Uint and type2.type in Uint:
        return _implicitly_convertible_to_for_uints(type1, type2)
    if type1.type in Int and type2.type in Int:
        return _implicitly_convertible_to_for_ints(type1, type2)
    if (
        type1.type != "bytes"
        and type2.type != "bytes"
        and type1.type in Byte
        and type2.type in Byte
    ):
        return _implicitly_convertible_to_for_bytes(type1, type2)
    return False


def _implicitly_convertible_to_for_bytes(type1: ElementaryType, type2: ElementaryType) -> bool:
    """
    Returns True if type1 may be implicitly converted to type2 assuming they are both bytes.
    """
    assert type1.type in Byte and type2.type in Byte
    assert type1.size is not None
    assert type2.size is not None

    return type1.size <= type2.size


def _implicitly_convertible_to_for_addresses(type1: ElementaryType, type2: ElementaryType) -> bool:
    """
    Returns True if type1 may be implicitly converted to type2 assuming they are both addresses.
    """
    assert type1.type == "address" and type2.type == "address"
    # payable attribute to be implemented; for now, always return True
    return True


def _implicitly_convertible_to_for_ints(type1: ElementaryType, type2: ElementaryType) -> bool:
    """
    Returns True if type1 may be implicitly converted to type2 assuming they are both ints.
    """
    assert type1.type in Int and type2.type in Int
    assert type1.size is not None
    assert type2.size is not None

    return type1.size <= type2.size


def _implicitly_convertible_to_for_uints(type1: ElementaryType, type2: ElementaryType) -> bool:
    """
    Returns True if type1 may be implicitly converted to type2 assuming they are both uints.
    """
    assert type1.type in Uint and type2.type in Uint
    assert type1.size is not None
    assert type2.size is not None

    return type1.size <= type2.size


def _implicitly_convertible_to_for_contracts(contract1: Contract, contract2: Contract) -> bool:
    """
    Returns True if contract1 may be implicitly converted to contract2.
    """
    return contract1 == contract2 or contract2 in contract1.inheritance


class IncorrectUsingFor(AbstractDetector):
    """
    Detector for incorrect using-for statement usage.
    """

    ARGUMENT = "incorrect-using-for"
    HELP = "Detects using-for statement usage when no function from a given library matches a given type"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#incorrect-usage-of-using-for-statement"

    WIKI_TITLE = "Incorrect usage of using-for statement"
    WIKI_DESCRIPTION = (
        "In Solidity, it is possible to use libraries for certain types, by the `using-for` statement "
        "(`using <library> for <type>`). However, the Solidity compiler doesn't check whether a given "
        "library has at least one function matching a given type. If it doesn't, such a statement has "
        "no effect and may be confusing. "
    )

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
    ```solidity
    library L {
        function f(bool) public pure {}
    }

    using L for uint;
    ```
    Such a code will compile despite the fact that `L` has no function with `uint` as its first argument."""
    # endregion wiki_exploit_scenario
    WIKI_RECOMMENDATION = (
        "Make sure that the libraries used in `using-for` statements have at least one function "
        "matching a type used in these statements. "
    )

    def _append_result(
        self, results: list[Output], uf: UsingForTopLevel, type_: Type, library: Contract
    ) -> None:
        info: DETECTOR_INFO = [
            f"using-for statement at {uf.source_mapping} is incorrect - no matching function for {type_} found in ",
            library,
            ".\n",
        ]
        res = self.generate_result(info)
        results.append(res)

    def _detect(self) -> list[Output]:
        results: list[Output] = []

        for uf in self.compilation_unit.using_for_top_level:
            # UsingForTopLevel.using_for is a dict with a single entry, which is mapped to a list of functions/libraries
            # the following code extracts the type from using-for and skips using-for statements with functions
            type_ = list(uf.using_for.keys())[0]
            for lib_or_fcn in uf.using_for[type_]:
                # checking for using-for with functions is already performed by the compiler; we only consider libraries
                if isinstance(lib_or_fcn, UserDefinedType):
                    lib_or_fcn_type = lib_or_fcn.type
                    if (
                        isinstance(type_, Type)
                        and isinstance(lib_or_fcn_type, Contract)
                        and not _is_correctly_used(type_, lib_or_fcn_type)
                    ):
                        self._append_result(results, uf, type_, lib_or_fcn_type)

        return results



"""
Event module
"""

from slither.core.variables.event_variable import EventVariable
from slither.vyper_parsing.variables.event_variable import EventVariableVyper
from slither.core.declarations.event import Event
from slither.vyper_parsing.ast.types import AnnAssign, Pass


from slither.vyper_parsing.ast.types import EventDef


class EventVyper:
    """
    Event class
    """

    def __init__(self, event: Event, event_def: EventDef) -> None:
        self._event = event
        self._event.name = event_def.name
        self._elemsNotParsed = event_def.body

    def analyze(self, contract) -> None:
        for elem_to_parse in self._elemsNotParsed:
            if not isinstance(elem_to_parse, AnnAssign):
                assert isinstance(elem_to_parse, Pass)
                continue

            elem = EventVariable()

            elem.set_offset(elem_to_parse.src, self._event.contract.compilation_unit)
            event_parser = EventVariableVyper(elem, elem_to_parse)
            event_parser.analyze(contract)

            self._event.elems.append(elem)

        self._elemsNotParsed = []
